import json
from functools import wraps

import pandas as pd
from platformdirs import user_data_path


# helper functions for pandas dataframes
def separate(data, col, into=None, sep=" ", **kwargs):
    into = into or col
    new_data = data.copy()
    new_data[into] = data[col].str.split(sep, **kwargs)
    return new_data


def cache_file(func):
    """
    Decorator for caching mapping dictionaries in ~/.local/share/gemsembler
    """

    @wraps(func)
    def wrapper_decorator(*args, **kwargs):
        # Create the directory holding
        cache_path = user_data_path("gemsembler", ensure_exists=True)
        cache_path /= func.__name__ + ".json"

        if cache_path.exists():
            with open(cache_path, "r") as fh:
                data = json.load(fh)
        else:
            data = func(*args, **kwargs)
            with open(cache_path, "w") as fh:
                json.dump(data, fh)
        return data

    return wrapper_decorator


def download_db(url, cache_name=None, **kwargs):
    """
    Function to download the data needed for conversion. Caches the data in
    ~/.local/share/gemsembler.
    """
    # Either get filename for cache file from input arg or url
    cache_name = cache_name or url.rsplit("/", 1)[-1]

    # Define the path to where to cache the data
    cache_path = user_data_path("gemsembler", ensure_exists=True) / cache_name

    # If the cache file exists open it, otherwise download the data
    if cache_path.exists():
        data = pd.read_csv(cache_path, sep="\t", **kwargs)
    else:
        print(f"Downloading {cache_name} from {url}")
        data = pd.read_csv(url, sep="\t", **kwargs)
        data.to_csv(cache_path, sep="\t", index=False)
    return data


def process_bigg(data, metabolites=False):
    """
    Function to process the BiGG database and get the mapping between old and
    new BiGG ids.
    """
    return (
        data.dropna(subset="old_bigg_ids")
        .copy()
        .pipe(separate, "old_bigg_ids", sep="; ")
        .explode(column="old_bigg_ids")
        .get(["old_bigg_ids", "bigg_id"])
        .assign(
            bigg_id=lambda x: (
                x.bigg_id.str.replace("_[a-z]+$", "", regex=True)
                if metabolites
                else x.bigg_id
            )
        )
        .drop_duplicates()
        .groupby("old_bigg_ids", group_keys=False)["bigg_id"]
        .apply(list)
        .to_dict()
    )


@cache_file
def get_old_bigg_m():
    df_bigg_m = download_db(
        "http://bigg.ucsd.edu/static/namespace/bigg_models_metabolites.txt",
        "bigg_models_metabolites.txt.gz",
    )
    return df_bigg_m.pipe(process_bigg, metabolites=True)


@cache_file
def get_old_bigg_r():
    df_bigg_r = download_db(
        "http://bigg.ucsd.edu/static/namespace/bigg_models_reactions.txt",
        "bigg_models_reactions.txt.gz",
    )
    return df_bigg_r.pipe(process_bigg)


def process_modelseed(data):
    return (
        data[["id", "aliases"]]
        .set_index("id")["aliases"]
        .str.extract(r"BiGG: (.*?)\|")[0]
        .str.split("; ")
        .reset_index(name="bigg_ids")
        .rename(columns={"id": "seed_ids"})
        .set_index("seed_ids")
        .get("bigg_ids")
        .dropna()
        .to_dict()
    )


@cache_file
def get_seed_orig_m():
    df_modelseed_m = download_db(
        (
            "https://github.com/ModelSEED/ModelSEEDDatabase/"
            "raw/master/Biochemistry/compounds.tsv"
        ),
        "compounds.tsv.gz",
    )
    return df_modelseed_m.pipe(process_modelseed)


@cache_file
def get_seed_orig_r():
    df_modelseed_r = download_db(
        (
            "https://github.com/ModelSEED/ModelSEEDDatabase/"
            "raw/master/Biochemistry/reactions.tsv"
        ),
        "reactions.tsv.gz",
    )
    return df_modelseed_r.pipe(process_modelseed)


def process_with_metanetx(data, db_name, repl_regex):
    """Get a mapping between `db_name` and bigg using metanetx db."""
    return (
        data.query("source.str.startswith(@db_name) or source.str.startswith('bigg')")
        .copy()
        .reset_index(drop=True)
        .assign(source=lambda x: x.source.str.replace(repl_regex, ":", regex=True))
        .pipe(separate, "source", sep=":", into=["db", "db_id"], expand=True)
        .drop_duplicates(subset=["db", "db_id"])
        .pivot_table(index="ID", columns="db", values="db_id", aggfunc=",".join)
        .dropna()
        .assign(other_db=lambda x: x[db_name].str.split(","))
        .drop(columns=db_name)
        .explode("other_db")
        .set_index("other_db")
        .get("bigg")
        .str.split(",")
        .to_dict()
    )


@cache_file
def get_seed_addit_m():
    df_metanetx_m = download_db(
        "https://www.metanetx.org/cgi-bin/mnxget/mnxref/chem_xref.tsv",
        "chem_xref.tsv.gz",
        comment="#",
        names=["source", "ID", "description"],
    )
    return df_metanetx_m.pipe(
        process_with_metanetx, "seed", r"(\.compound|M|\.metabolite):(M_)?"
    )


@cache_file
def get_seed_addit_r():
    df_metanetx_r = download_db(
        "https://www.metanetx.org/cgi-bin/mnxget/mnxref/reac_xref.tsv",
        "reac_xref.tsv.gz",
        comment="#",
        names=["source", "ID", "description"],
    )
    return df_metanetx_r.pipe(process_with_metanetx, "seed", r"(\.reaction|R|):(R_)?")


def process_metanetx(data, repl_regex):
    """Get a mapping between metanetx and bigg"""
    return (
        data.query("source.str.startswith('bigg') and ID != 'EMPTY'")
        .assign(source=lambda x: x.source.str.replace(repl_regex, "", regex=True))
        .rename(columns={"source": "bigg", "ID": "mnx"})
        .drop_duplicates(subset="bigg")
        .drop(columns="description")
        .groupby("mnx")["bigg"]
        .apply(list)
        .to_dict()
    )


@cache_file
def get_mnx_m():
    df_metanetx_m = download_db(
        "https://www.metanetx.org/cgi-bin/mnxget/mnxref/chem_xref.tsv",
        "chem_xref.tsv.gz",
        comment="#",
        names=["source", "ID", "description"],
    )
    return df_metanetx_m.pipe(
        process_metanetx, r"bigg(\.compound|M|\.metabolite):(M_)?"
    )


@cache_file
def get_mnx_r():
    df_metanetx_r = download_db(
        "https://www.metanetx.org/cgi-bin/mnxget/mnxref/reac_xref.tsv",
        "reac_xref.tsv.gz",
        comment="#",
        names=["source", "ID", "description"],
    )
    return df_metanetx_r.pipe(process_metanetx, r"bigg(\.reaction|R|):(R_)?")


@cache_file
def get_kegg_m():
    df_metanetx_m = download_db(
        "https://www.metanetx.org/cgi-bin/mnxget/mnxref/chem_xref.tsv",
        "chem_xref.tsv.gz",
        comment="#",
        names=["source", "ID", "description"],
    )
    return df_metanetx_m.pipe(
        process_with_metanetx,
        "kegg",
        r"(\.compound|\.drug|\.metabolite|\.glycan|[CDGM]):(M_)?",
    )


@cache_file
def get_kegg_r():
    df_metanetx_r = download_db(
        "https://www.metanetx.org/cgi-bin/mnxget/mnxref/reac_xref.tsv",
        "reac_xref.tsv.gz",
        comment="#",
        names=["source", "ID", "description"],
    )
    return df_metanetx_r.pipe(process_with_metanetx, "kegg", r"(\.reaction|R|):(R_)?")


def get_BiGG_lists(metabolites: bool):
    if metabolites:
        bigg_data = download_db(
            "http://bigg.ucsd.edu/static/namespace/bigg_models_metabolites.txt",
            "bigg_models_metabolites.txt.gz",
        )
        return set(bigg_data.get("universal_bigg_id"))
    else:
        bigg_data = download_db(
            "http://bigg.ucsd.edu/static/namespace/bigg_models_reactions.txt",
            "bigg_models_reactions.txt.gz",
        )
        return set(bigg_data.get("bigg_id"))


def remove_duplicates(data, leave_from_mixed_directions=True):
    duplicated = (
        data.copy()
        .rename(
            columns={"1metabolites": "2metabolites", "2metabolites": "1metabolites"}
        )
        .pipe(lambda x: pd.concat([x, data], ignore_index=True))
        .sort_values(
            ["1metabolites", "2metabolites", "models_number", "reaction"],
            ascending=[True, True, False, True],
        )
        .pipe(lambda x: x[x.duplicated(["1metabolites", "2metabolites"], keep=False)])
    )

    reaction_list = duplicated["reaction"].unique().tolist()
    reaction_list_wo_dupl_metabolites = (
        duplicated.drop_duplicates(["1metabolites", "2metabolites"], keep="first")[
            "reaction"
        ]
        .unique()
        .tolist()
    )
    if leave_from_mixed_directions:
        reaction_list = set(reaction_list) - set(reaction_list_wo_dupl_metabolites)
    return data[~data["reaction"].isin(reaction_list)]


def get_bigg_network(path_to_dbs=None, leave_from_mixed_directions=True):
    """
    Getting dictionary for BiGG topology: unique reaction equation as key and
    id as value.  Maybe writing down a table with reaction ids unique reaction
    equations with metabolites sorted for the whole BiGG database. If reaction
    equations are duplicated one used in most amount of models selected.
    """

    bigg_database_r = download_db(
        "http://bigg.ucsd.edu/static/namespace/bigg_models_reactions.txt",
        "bigg_models_reactions.txt.gz",
    )

    r_connections = (
        bigg_database_r["reaction_string"]
        .str.replace(r"(\d+\.\d*(e-)?\d*|\d+e-\d*)|\+", "", regex=True)
        .apply(lambda x: f" {x} ")
        .str.split("<->", n=1, expand=True)
        .rename(columns={0: "1metabolites", 1: "2metabolites"})
        # Sort the lists of metabolites in both columns
        .map(lambda x: " ".join(sorted(x.split())))
        .assign(
            reaction=bigg_database_r["bigg_id"],
            models_number=bigg_database_r["model_list"].str.split().apply(len),
        )
        # Sort the dataframe by the metabolites
        .sort_values(
            ["1metabolites", "2metabolites", "models_number"],
            ascending=[True, True, False],
        )
        # Remove duplicates based on metabolites
        .drop_duplicates(["1metabolites", "2metabolites"], keep="first")
        # Remove duplicates based on reactions
        .pipe(remove_duplicates, leave_from_mixed_directions)
        # Add an equation column
        .assign(
            equation=lambda x: (
                x[["1metabolites", "2metabolites"]].apply(
                    lambda row: "<->".join(sorted(row)), axis=1
                )
            )
        )
        .set_index("equation")
        .get("reaction")
        .to_dict()
    )

    return r_connections



import argparse
import pickle
import tempfile
from ast import literal_eval
from contextlib import contextmanager
from copy import deepcopy
from os import chdir, getcwd
from pathlib import Path

import h5py
import metquest
from cobra import Model
from cobra.io import read_sbml_model, write_sbml_model


# Context manager to get around metquest chdir internally
@contextmanager
def tmp_cwd():
    oldpwd = getcwd()
    try:
        yield
    finally:
        chdir(oldpwd)


def write_mq_data(mq_data: dict, filename: Path):
    """
    Function to save output from metquest in hdf5 format.
    HDF5 format was chosen due to large size of the output from metquest, which
    caused crashes when reading the data.

    Parameters
    ----------
    mq_data : dict
        Dictionary containing output from metquest
    filename : str
        path to output file
    """

    # Open file handle with h5py
    with h5py.File(filename, "w") as fh:
        # Save the metadata at root of the hierarchy in the file
        fh.create_dataset("all_synthesized", data=list(mq_data["all_synthesized"]))
        fh.create_dataset("max_synthesized", data=list(mq_data["max_synthesized"]))
        fh.create_dataset("max_paths_length", data=mq_data["max_paths_length"])
        fh.create_dataset("tag", data=mq_data["tag"])

        # Save the dictionary containing the mapping between metquest internal
        # ids and original ids
        name_map = fh.create_group("name_map")
        for mq_id, orig_id in mq_data["name_map"].items():
            name_map[mq_id] = orig_id

        # HDF5 format does not support saving dictionaries, so they need to
        # converted to strings before being saved. To ease memory footprint, we
        # stringify data at path length level, i.e.
        # /path_type/mb_id/path_length/<list_of_paths>
        # Applies for both linear and circular paths

        # Save linear linear paths
        linear_paths = fh.create_group("linear_paths")
        for mb_id, path_lengths in mq_data["linear_paths"].items():
            linear_paths.create_group(mb_id)
            for path_length, path in path_lengths.items():
                linear_paths[mb_id][str(path_length)] = str(path)

        # Save the circular paths
        circular_paths = fh.create_group("circular_paths")
        for mb_id, path_lengths in mq_data["circular_paths"].items():
            circular_paths.create_group(mb_id)
            for path_length, path in path_lengths.items():
                circular_paths[mb_id][str(path_length)] = str(path)


class MQData(object):
    def __init__(self, filename):
        # Save the filename as private variable to prevent the user from
        # accidentally overwriting it. Get and store metadata from the file.
        self.__filename = filename
        with h5py.File(self.__filename, "r") as fh:
            self.__max_paths_length = fh["max_paths_length"][()]
            self.__tag = fh["tag"][()].decode()

            self.__name_map = {
                mq_id: orig_id[()].decode() for mq_id, orig_id in fh["name_map"].items()
            }

            self.__rev_name_map = {
                orig_id: mq_id for mq_id, orig_id in self.__name_map.items()
            }

    # To minimise memory footprint, only load this data when requested by user
    @property
    def all_synthesized(self):
        with h5py.File(self.__filename, "r") as fh:
            return {x.decode() for x in fh["all_synthesized"]}

    # To minimise memory footprint, only load this data when requested by user
    @property
    def max_synthesized(self):
        with h5py.File(self.__filename, "r") as fh:
            return {x.decode() for x in fh["max_synthesized"]}

    # Property to prevent the user from overwriting it
    @property
    def name_map(self):
        return self.__name_map

    # Property to prevent the user from overwriting it
    @property
    def reverse_name_map(self):
        return self.__rev_name_map

    # Property to prevent the user from overwriting it
    @property
    def max_paths_length(self):
        return self.__max_paths_length

    # Property to prevent the user from overwriting it
    @property
    def tag(self):
        return self.__tag

    # Function to get available keys at the given level
    def get_keys(self, path_type, mb_id=None):
        with h5py.File(self.__filename, "r") as fh:
            if mb_id is not None:
                return list(fh[f"{path_type}/{mb_id}"].keys())
            else:
                return list(fh[path_type].keys())

    # Wrapper function around get_keys for better ux
    def get_ids(self, path_type):
        return self.get_keys(path_type)

    # Wrapper function around get_keys for better ux
    def get_paths_lengths(self, path_type, mb_id):
        return self.get_keys(path_type, mb_id)

    # Function to get all paths given path type, metabolite id and path length
    def __get_by_path_length(self, path_type, mb_id, path_length, convert=False):
        with h5py.File(self.__filename) as fh:
            paths = literal_eval(fh[f"{path_type}/{mb_id}/{path_length}"][()].decode())
            if convert:
                paths = [{self.__name_map[_id] for _id in path} for path in paths]
            return paths

    # Function to get all paths for all path lengths given path type and
    # metabolite id
    def __get_by_mb_id(self, path_type, mb_id, convert=False):
        return {
            path_len: self.__get_by_path_length(path_type, mb_id, path_len, convert)
            for path_len in self.get_paths_lengths(path_type, mb_id)
        }

    # Defining __getitem__ method to be able to access the data with
    # mq_data[<path_type>, <mb_id>, <path_length>]
    def __getitem__(self, key):
        # If only path type is given, return all paths for all metabolites
        if isinstance(key, str):
            return {
                mb_id: self.__get_by_mb_id(key, mb_id) for mb_id in self.get_ids(key)
            }
        # If more arguments were specified
        elif hasattr(key, "__iter__"):
            # Return all paths for all path lengths given metabolite id
            if len(key) == 2:
                return self.__get_by_mb_id(*key)
            # Return all paths for given path type, metabolite id and path length
            elif len(key) == 3:
                return self.__get_by_path_length(*key)
            else:
                raise ValueError(f"Too many args: {len(key)}")

    def _subset_paths(
        self, path_type, mb_id, path_lengths=None, len_diversity=None, map_back=True
    ):
        # Get path lengths if not defined, otherwise check the correct type
        if path_lengths is None:
            # path lengths are stored as strings in the file, so needs
            # conversion back to ints and sorting
            path_lengths = sorted(
                [int(x) for x in self.get_paths_lengths(path_type, mb_id)]
            )
        elif isinstance(path_lengths, int):
            path_lengths = [path_lengths]
        elif isinstance(path_lengths, list):
            pass
        else:
            raise ValueError("path_lengths can only be: None, int or list")

        # Subset path lengths based on len_diversity arg if given
        if len_diversity is not None:
            len_diversity = min(len(path_lengths), len_diversity)
            path_lengths = path_lengths[:len_diversity]

        # Return all paths for given path lengths (if specified)
        # In case paths are there: length bigger than 0
        # otherwise return empty dict
        if 0 in path_lengths:
            return {0: []}
        else:
            return {
                path_len: self.__get_by_path_length(
                    path_type, mb_id, path_len, map_back
                )
                for path_len in path_lengths
            }

    # Wrapper function around _subset_paths to allow for getting all paths for
    # all metabolites
    def subset_paths(
        self,
        path_type,
        mb_id=None,
        path_lengths=None,
        len_diversity=None,
        map_back=True,
    ):
        if mb_id is None:
            return {
                self._subset_paths(
                    path_type, mb_id, path_lengths, len_diversity, map_back
                )
                for mb_id in self.get_ids(path_type)
            }
        else:
            return self._subset_paths(
                path_type, mb_id, path_lengths, len_diversity, map_back
            )

    # metquest is not consistent with metabolite ids, hence we get all
    # possibilites
    def _get_alt_mb_ids(self, mb_id):
        mb_id_tag = self.__tag + " " + mb_id
        mb_id_name_map = self.__rev_name_map.get(mb_id, "")
        return mb_id_tag, mb_id_name_map

    # Function to check which of the given metabolites is present in set of ids
    # for given path type. args is a list of possible metabolite ids.
    def _check_mb_id(self, path_type, *args):
        pos_ids = self.get_ids(path_type)
        for mb_id in args:
            if mb_id in pos_ids:
                return mb_id

    def get_all_paths(self, metabolites, len_diversity=3):
        met_paths = {}
        comments = {}
        all_syn = self.all_synthesized
        max_syn = self.max_synthesized

        for met in metabolites:
            # metquest adds model id to the beginning (sometimes - TODO)
            # Get all possible ways metquest can give ids to metabolites
            met_tag, met_name_map = self._get_alt_mb_ids(met)

            # if metabolite is not in all_synthesized list, then it cannot be
            # synthesized
            if (
                (met_tag not in all_syn)
                and (met not in all_syn)
                and (met_name_map not in all_syn)
            ):
                comments[met] = "can not be synthesized"
                met_paths[met] = {}

            # if metabolite is not in X_synthesized then it cannot be
            # synthesized with max path length of X
            elif (
                (met_tag not in max_syn)
                and (met not in max_syn)
                and (met_name_map not in max_syn)
            ):
                comment = f"can not be synthesized with max {self.max_paths_length} length path"
                comments[met] = comment
                met_paths[met] = {}
            else:
                met_found = self._check_mb_id(
                    "linear_paths", met_tag, met, met_name_map
                )
                if met_found is not None:
                    # Check linear paths for given metabolite. Get linear and
                    # circular paths from metquest output for a given metabolite
                    linear_paths = self._subset_paths(
                        "linear_paths", met_found, len_diversity=len_diversity
                    )
                    if linear_paths == {0: []}:
                        comments[met] = "Is in the medium, no need to synthesys"
                        met_paths[met] = {}
                    elif isinstance(linear_paths, str):
                        comments[met] = linear_paths
                    elif isinstance(linear_paths, dict):
                        comments[met] = "Found in linear paths"
                        met_paths[met] = linear_paths
                else:
                    met_found = self._check_mb_id(
                        "circular_paths", met_tag, met, met_name_map
                    )
                    if met_found is not None:
                        # Check circular paths for given metabolite
                        circular_paths = self._subset_paths(
                            "circular_paths", met_found, len_diversity=len_diversity
                        )
                        if circular_paths == {0: []}:
                            comments[met] = "Is in the medium, no need to synthesys"
                            met_paths[met] = {}
                        elif isinstance(circular_paths, str):
                            comments[met] = circular_paths
                        elif isinstance(circular_paths, dict):
                            comments[met] = "Found in circular paths"
                            met_paths[met] = circular_paths
                    else:
                        # Otherwise check ...
                        maybe_synthesised = self.all_synthesized - self.max_synthesized
                        if maybe_synthesised:  # if non-empty
                            comment = (
                                "Problematic: can be synthesized "
                                "but does not have paths"
                            )
                        else:
                            comment = (
                                "Maybe can not be synthesized with max "
                                f"{self.max_paths_length} length path, because "
                                "all_synthesized not bigger than max_synthesized"
                            )
                        comments[met] = comment
                        met_paths[met] = {}

        return met_paths, comments


def check_which_met_id_found(path_data: dict, *args):
    pos_ids = path_data.keys()
    for mb_id in args:
        if mb_id in pos_ids:
            return mb_id


def get_paths_for_met(path_data: dict, met: str, len_diversity: int, name_map: dict):
    len_diversity = min(len(path_data[met]), len_diversity)
    path_lengths = list(path_data[met].keys())[:len_diversity]
    # Return all paths for given path lengths (if specified)
    # In case paths are there: length bigger than 0, otherwise return empty dict
    if 0 in path_lengths:
        return {0: []}
    met_paths = {}
    for path_len in path_lengths:
        paths = path_data[met][path_len]
        new_paths = []
        for path in paths:
            new_paths.append({name_map[r_id] for r_id in path})
        met_paths[path_len] = new_paths
    return met_paths


def get_shortest_paths_data(mq_out_data: dict, metabolites: list, len_diversity=3):
    met_paths = {}
    comments = {}
    all_syn = mq_out_data["all_synthesized"]
    max_syn = mq_out_data["max_synthesized"]

    for met in metabolites:
        # Get all possible ways metquest can give ids to metabolites
        met_tag = mq_out_data["tag"] + " " + met
        reverse_name_map = {
            orig_id: mq_id for mq_id, orig_id in mq_out_data["name_map"].items()
        }
        met_name_map = reverse_name_map.get(met, "")

        # if metabolite is not in all_synthesized list, then it cannot be
        # synthesized
        if (
            (met_tag not in all_syn)
            and (met not in all_syn)
            and (met_name_map not in all_syn)
        ):
            comments[met] = "can not be synthesized"
            met_paths[met] = {}

        # if metabolite is not in X_synthesized then it cannot be
        # synthesized with max path length of X
        elif (
            (met_tag not in max_syn)
            and (met not in max_syn)
            and (met_name_map not in max_syn)
        ):
            comment = f"can not be synthesized with max {mq_out_data['max_paths_length']} length path"
            comments[met] = comment
            met_paths[met] = {}
        else:
            met_found = check_which_met_id_found(
                mq_out_data["linear_paths"], met, met_tag, met_name_map
            )
            if met_found is not None:
                # Check linear paths for given metabolite. Get linear and
                # circular paths from metquest output for a given metabolite
                linear_paths = get_paths_for_met(
                    mq_out_data["linear_paths"],
                    met_found,
                    len_diversity=len_diversity,
                    name_map=mq_out_data["name_map"],
                )
                if linear_paths == {0: []}:
                    comments[met] = "Is in the medium, no need to synthesys"
                    met_paths[met] = {}
                elif isinstance(linear_paths, dict):
                    comments[met] = "Found in linear paths"
                    met_paths[met] = linear_paths
            else:
                c_met_found = check_which_met_id_found(
                    mq_out_data["circular_paths"], met, met_tag, met_name_map
                )
                if c_met_found is not None:
                    # Check circular paths for given metabolite
                    circular_paths = get_paths_for_met(
                        mq_out_data["circular_paths"],
                        c_met_found,
                        len_diversity=len_diversity,
                        name_map=mq_out_data["name_map"],
                    )
                    if circular_paths == {0: []}:
                        comments[met] = "Is in the medium, no need to synthesys"
                        met_paths[met] = {}
                    elif isinstance(circular_paths, dict):
                        comments[met] = "Found in circular paths"
                        met_paths[met] = circular_paths
                else:
                    # Otherwise check ...
                    maybe_synthesised = all_syn - max_syn
                    if maybe_synthesised:  # if non-empty
                        comment = (
                            "Problematic: can be synthesized " "but does not have paths"
                        )
                    else:
                        comment = (
                            "Maybe can not be synthesized with max "
                            f"{mq_out_data['max_paths_length']} length path, because "
                            "all_synthesized not bigger than max_synthesized"
                        )
                    comments[met] = comment
                    met_paths[met] = {}
    return {"paths": met_paths, "comments": comments}


def preprocess_medium(
    tag: str, nutritional_sources: [str], other_medium: [str], cofactors: [str],
):
    if cofactors is None:
        cofactors = [
            "co2",
            "hco3",
            "pi",
            "ppi",
            "ACP",
            "atp",
            "adp",
            "amp",
            "nad",
            "nadh",
            "nadp",
            "nadph",
            "coa",
            "cmp",
            "cdp",
            "ctp",
            "gmp",
            "gdp",
            "gtp",
            "ump",
            "udp",
            "utp",
            "fadh2",
            "fad",
            "q8",
            "q8h2",
            "mqn8",
            "mql8",
            "mqn6",
            "mql6",
            "thf",
        ]
    # TODO: Why is `_c` added to each metabolite id?
    cofactors_c = [c + "_c" for c in cofactors]
    other_medium_c = [m + "_c" for m in other_medium]
    nutritional_sources_c = [n + "_c" for n in nutritional_sources]
    seed_metabolites = nutritional_sources_c + other_medium_c + cofactors_c
    seed_metabolites_tag = set([tag + " " + s for s in seed_metabolites])
    return seed_metabolites_tag, cofactors_c, other_medium_c


def remove_cofactors_from_model(
    model: Model, medium_wo_cof_c: [str], cofactors_c: [str]
):
    r_to_remove = []
    met_c = medium_wo_cof_c + cofactors_c
    for r in model.reactions:
        p = True
        if (r.lower_bound == 0) and (r.upper_bound > 0):
            for react in r.reactants:
                if react.id not in met_c:
                    p = False
        elif (r.lower_bound < 0) and (r.upper_bound == 0):
            for pro in r.products:
                if pro.id not in met_c:
                    p = False
        else:
            pp = True
            for react in r.reactants:
                if react.id not in met_c:
                    pp = False
            if not pp:
                pp = True
                for pro in r.products:
                    if pro.id not in met_c:
                        pp = False
            p = pp
        if p:
            r_to_remove.append(r.id)

    model.remove_reactions(r_to_remove)
    return model


def run_metquest(
    path_to_models: str, seed_metabolites_tag: [str], number_of_xml, max_paths_length,
):
    graph, name_map = metquest.construct_graph.create_graph(
        path_to_models, number_of_xml
    )

    lowboundmet_pic, status_dict_pic, scope_pic = metquest.forward_pass(
        graph, seed_metabolites_tag
    )

    # BE CAREFULLY can crash computer: memory heavy and long. Better do on cluster.
    pathway_table, c_path, scope = metquest.find_pathways(
        graph, seed_metabolites_tag, max_paths_length
    )

    output = {
        "all_synthesized": scope_pic,
        "max_paths_length": max_paths_length,
        "max_synthesized": scope,
        "linear_paths": pathway_table,
        "circular_paths": c_path,
        "name_map": name_map,
    }
    return output, graph


def pathsfinding(
    path_to_model,
    output_dir,
    nutritional_sources: [str],
    other_medium: [str],
    cofactors=None,
    max_paths_length=40,
    number_of_xml=1,
    len_diversity=3,
    metabolites=None,
    **kwargs,
):
    # Convert paths to Path objects
    path_to_model = Path(path_to_model)
    model_name = path_to_model.stem
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    # Open the models
    model = read_sbml_model(path_to_model)
    tag = model.id

    # Alter metabolite ids
    seed_metabolites_tag, cofactors_c, other_medium_c = preprocess_medium(
        tag, nutritional_sources, other_medium, cofactors
    )

    # Remove cofactors from the model
    model_wo_cof = remove_cofactors_from_model(
        deepcopy(model), other_medium_c, cofactors_c
    )

    # Save the model without cofactors
    path_to_model_wo_cof = output_dir / (model_name + "_wo_cofactors.xml")
    write_sbml_model(model_wo_cof, path_to_model_wo_cof)

    # Make a tmp dir in which to run metquest
    with tempfile.TemporaryDirectory(dir=output_dir) as tmp_dir, tmp_cwd():
        # Make symbolic link to model
        tmp_model_path = Path(tmp_dir) / (model_name + ".xml")
        tmp_model_path.symlink_to(f"../{model_name}_wo_cofactors.xml")

        # Run metquest
        mq_output, graph = run_metquest(
            tmp_dir, seed_metabolites_tag, number_of_xml, max_paths_length,
        )

    # Add necessary parameters to output data
    mq_output.update({"tag": tag, "max_paths_length": max_paths_length})

    # Save the output
    write_mq_data(mq_output, output_dir / "metquest.h5")

    if metabolites is None:
        metabolites = [m.id for m in model_wo_cof.metabolites]
    shortest_paths = get_shortest_paths_data(mq_output, metabolites, len_diversity)

    with open(output_dir / "shortest_paths.pkl", "wb") as fh:
        pickle.dump(shortest_paths, fh)

    with open(output_dir / "graph.pkl", "wb") as fh:
        pickle.dump(graph, fh)



import math
import operator
import re
import networkx as nx
import pandas as pd
import seaborn as sns
from .creation import NewElement, SuperModel
from .comparison import getCoreGPR, getCoreConnections
from pyvis.network import Network
from copy import deepcopy

MET_NOT_INT_GLOBAL = {
    "h": 0,
    "h2o": 0,
    "h2": 0,
    "oh1": 0,
    "o2": 0,
    "co2": 0,
    "coa": 0,
    "ppi": 0,
    "pi": 0,
    "amp": 0,
    "adp": 0,
    "atp": 0,
    "cmp": 0,
    "cdp": 0,
    "ctp": 0,
    "gmp": 0,
    "gdp": 0,
    "gtp": 0,
    "ump": 0,
    "udp": 0,
    "utp": 0,
    "nad": 0,
    "nadh": 0,
    "nadp": 0,
    "nadph": 0,
    "dadp": 0,
    "damp": 0,
    "nh3": 0,
    "nh4": 0,
    "fadh2": 0,
    "fad": 0,
    "ac": 0,
    "accoa": 0,
    "h2s": 0,
    "HC00250": 0,
}


def get_color_palette(sources_number: int) -> dict:
    up = math.ceil(sources_number / 10)
    down = math.ceil(sources_number / 10)
    colors = sns.color_palette("bright").as_hex()
    out_colors = {
        "metabolites": sns.light_palette(
            colors[4], sources_number + up + down
        ).as_hex()[up:],
        "intersection_r": sns.light_palette(
            colors[3], sources_number + up + down
        ).as_hex()[up:],
        "single_path_r": sns.light_palette(
            colors[1], sources_number + up + down
        ).as_hex()[up:-down],
        "genes": sns.light_palette(colors[8], sources_number + up + down).as_hex()[
            up:-down
        ],
        "metConnect": sns.light_palette(colors[0], sources_number + up + down).as_hex()[
            up:-down
        ],
        "reactionsConnect": sns.light_palette(
            colors[9], sources_number + up + down
        ).as_hex()[up:-down],
        "interest": sns.light_palette(colors[2], sources_number + up + down).as_hex()[
            up:-down
        ],
        "notFound": sns.light_palette(colors[7], sources_number + up + down).as_hex()[
            up:-down
        ],
    }
    return out_colors


def custom_histplot(data, y, hue, **kwargs):
    unique_confidences = list(data["Confidence"].sort_values(ascending=False).unique())
    num_levels = len(unique_confidences)
    if data["Reactions/GPRs"].iloc[0] == "Reactions":
        base_color = sns.color_palette("bright")[1]
    else:
        base_color = sns.color_palette("bright")[8]
    palette_colors = sns.light_palette(base_color, n_colors=num_levels).as_hex()[::-1]
    palette = {level: color for level, color in zip(unique_confidences, palette_colors)}

    sns.histplot(
        data=data,
        stat="count",
        multiple="stack",
        y=y,
        hue=hue,
        kde=False,
        element="bars",
        palette=palette,
        **kwargs,
    )


def define_node_features(
    colordata: dict,
    pallitra: str,
    object: NewElement,
    n_letter: int,
    gene=False,
    and_as_solid=False,
    node_id=None,
    label=None,
) -> [str]:
    if not gene:
        col = colordata.get(pallitra)[object.in_models["models_amount"] - 1]
        short_name = "".join(
            [tmp[:n_letter] for tmp in object.in_models["models_list"]]
        )
        if node_id is None:
            node_id = object.id
        if label is None:
            label = object.id
        title = f"Confirmed by {object.in_models['models_amount']}:\n{short_name}"
    else:
        if not object.gene_reaction_rule["assembly"]:
            node_id = f"{object.id} NO GPR"
            label = "NO GPR"
            col = colordata.get("notFound")[object.in_models["models_amount"] - 1]
            title = "No genes found"
        else:
            for i in range(object.in_models["models_amount"], 0, -1):
                gpr = getCoreGPR(
                    object.gene_reaction_rule,
                    i,
                    operator.ge,
                    object.in_models["models_list"],
                    and_as_solid,
                )
                if gpr:
                    node_id = f"{object.id} GPR core {i}"
                    label = f"GPR core {i}"
                    col = colordata.get(pallitra)[i - 1]
                    title = (
                        f"GPR core {i}:\n{gpr[0]}\n"
                        f"GPR assembly:\n{object.gene_reaction_rule['assembly'][0]}"
                    )
                    break
    return [node_id, label, col, title]


def define_edge_features(
    colordata: dict,
    pallitra_forward: str,
    pallitra_reversed: str,
    mr: NewElement,
    colname_m: str,
    r: NewElement,
    colname_r: str,
    connections: str,
    n_letter: int,
) -> [[str, str, str, str]]:
    n = 0
    short_name = ""
    if r.lower_bound.get("assembly")[0] >= 0:
        for tmp in r.in_models["models_list"]:
            if mr in getattr(r, connections).get(tmp):
                n = n + 1
                short_name = short_name + tmp[:n_letter]
        if connections == "reactants":
            col = colordata.get(pallitra_forward)[n - 1]
            output = [[colname_m, colname_r, col, f"Confirmed by {n}:\n{short_name}"]]
        elif connections == "products":
            col = colordata.get(pallitra_reversed)[n - 1]
            output = [[colname_r, colname_m, col, f"Confirmed by {n}:\n{short_name}"]]
    elif r.upper_bound.get("assembly")[0] <= 0:
        for tmp in r.in_models["models_list"]:
            if mr in getattr(r, connections).get(tmp):
                n = n + 1
                short_name = short_name + tmp[:n_letter]
        if connections == "reactants":
            col = colordata.get(pallitra_reversed)[n - 1]
            output = [[colname_r, colname_m, col, f"Confirmed by {n}:\n{short_name}"]]
        elif connections == "products":
            col = colordata.get(pallitra_forward)[n - 1]
            output = [[colname_m, colname_r, col, f"Confirmed by {n}:\n{short_name}"]]
    else:
        f = 0
        b = 0
        short_name_f = ""
        short_name_b = ""
        for tmp in r.in_models["models_list"]:
            if mr in getattr(r, connections).get(tmp) and r.upper_bound.get(tmp)[0] > 0:
                f = f + 1
                short_name_f = short_name_f + tmp[:n_letter]
            if mr in getattr(r, connections).get(tmp) and r.lower_bound.get(tmp)[0] < 0:
                b = b + 1
                short_name_b = short_name_b + tmp[:n_letter]
        if connections == "reactants":
            col_f = colordata.get(pallitra_forward)[f - 1]
            col_b = colordata.get(pallitra_forward)[b - 1]
            output = [
                [colname_m, colname_r, col_f, f"Confirmed by {f}:\n{short_name_f}"],
                [colname_r, colname_m, col_b, f"Confirmed by {b}:\n{short_name_b}"],
            ]
        elif connections == "products":
            col_f = colordata.get(pallitra_reversed)[f - 1]
            col_b = colordata.get(pallitra_reversed)[b - 1]
            output = [
                [colname_r, colname_m, col_f, f"Confirmed by {f}:\n{short_name_f}"],
                [colname_m, colname_r, col_b, f"Confirmed by {b}:\n{short_name_b}"],
            ]
    return output


def get_pyvis_from_nx(g, directed, size, width=1920, height=1080):
    pyvis_graph = Network(
        width=f"{width}px", height=f"{height}px", directed=directed, notebook=False,
    )
    pyvis_graph.from_nx(g)
    for n in pyvis_graph.nodes:
        n["size"] = size
        n["font"] = {"size": size}
    return pyvis_graph


def draw_notconv_biomass(
    supermodel: SuperModel,
    output_name: str,
    directed=True,
    n_letter=None,
    wid=1920,
    hei=1080,
    size=25,
):
    if not output_name.endswith(".html"):
        raise ValueError(
            "Output file for the plot is of wrong format. Please use html file name."
        )
    if n_letter is None:
        n_letter = supermodel.get_short_name_len()
    g = nx.DiGraph()
    biomass_r = supermodel.reactions.notconverted.get("Biomass")
    reactants = {}
    products = {}
    for typ in supermodel.sources:
        for rea in biomass_r.reactants.get(typ):
            if rea.id not in reactants.keys():
                reactants.update({rea.id: f"{rea.id}\n{typ[:n_letter]}"})
            else:
                if rea.id in rea.annotation.get(typ):
                    reactants.update(
                        {rea.id: f"{reactants.get(rea.id)}{typ[:n_letter]}"}
                    )
        for pro in biomass_r.products.get(typ):
            if pro.id not in products.keys():
                products.update({pro.id: f"{pro.id}\n{typ[:n_letter]}"})
            else:
                products.update({pro.id: f"{products.get(pro.id)}{typ[:n_letter]}"})
    g.add_node(biomass_r.id, shape="box")
    for r in reactants.values():
        g.add_node(r, shape="o")
        g.add_edge(r, biomass_r.id, font_color="black")
    for p in products.values():
        g.add_node(p, shape="o")
        g.add_edge(biomass_r.id, p, font_color="black")

    pyvis_graph = get_pyvis_from_nx(g, directed, size, wid, hei)
    pyvis_graph.write_html(output_name, notebook=False)
    return g


def draw_one_known_pathway(
    supermodel: SuperModel,
    pathway: dict,
    output_name: str,
    additional_met=False,
    genes=True,
    and_as_solid=False,
    directed=True,
    met_not_int=None,
    n_letter=None,
    wid=1920,
    hei=1080,
    size=25,
):
    if not output_name.endswith(".html"):
        raise ValueError(
            "Output file for the plot is of wrong format. Please use html file name."
        )
    if met_not_int is None:
        met_not_int = deepcopy(MET_NOT_INT_GLOBAL)
    color_brewer = get_color_palette(len(supermodel.sources))
    if n_letter is None:
        n_letter = supermodel.get_short_name_len()
    g = nx.DiGraph()
    path_met = []
    for value in pathway.values():
        for v in value:
            path_met.append(v[0])
            path_met.append(v[1])
    path_met = list(set(path_met))
    for r_id in pathway.keys():
        if r_id in supermodel.reactions.assembly.keys():
            r = supermodel.reactions.assembly.get(r_id)
            colname_r = define_node_features(
                color_brewer, "single_path_r", r, n_letter,
            )
            g.add_node(
                colname_r[0],
                label=colname_r[1],
                shape="box",
                color=colname_r[2],
                title=colname_r[3],
            )
            for rea in r.reactants.get("assembly"):
                tmp_rea = re.sub("_([cep])$", "", rea.id)
                if rea.id in path_met:
                    colname_rea = define_node_features(
                        color_brewer, "metabolites", rea, n_letter,
                    )
                    rea_edge = define_edge_features(
                        color_brewer,
                        "single_path_r",
                        "metabolites",
                        rea,
                        colname_rea[1],
                        r,
                        colname_r[1],
                        "reactants",
                        n_letter,
                    )
                    g.add_node(
                        colname_rea[0],
                        shape="o",
                        label=colname_rea[1],
                        color=colname_rea[2],
                        title=colname_rea[3],
                    )
                    for e in rea_edge:
                        g.add_edge(
                            e[0], e[1], color=e[2], font_color="black",
                        )
                elif additional_met:
                    if tmp_rea not in met_not_int.keys():
                        colname_rea = define_node_features(
                            color_brewer, "metConnect", rea, n_letter,
                        )
                        rea_edge = define_edge_features(
                            color_brewer,
                            "single_path_r",
                            "metabolites",
                            rea,
                            colname_rea[1],
                            r,
                            colname_r[1],
                            "reactants",
                            n_letter,
                        )
                    else:
                        colname_rea = define_node_features(
                            color_brewer,
                            "metConnect",
                            rea,
                            n_letter,
                            node_id=f"{tmp_rea} {met_not_int[tmp_rea]}",
                            label=tmp_rea,
                        )
                        rea_edge = define_edge_features(
                            color_brewer,
                            "single_path_r",
                            "metabolites",
                            rea,
                            f"{tmp_rea} {met_not_int[tmp_rea]}",
                            r,
                            colname_r[1],
                            "reactants",
                            n_letter,
                        )
                        met_not_int[tmp_rea] = met_not_int[tmp_rea] + 1
                    g.add_node(
                        colname_rea[0],
                        shape="o",
                        label=colname_rea[1],
                        color=colname_rea[2],
                        title=colname_rea[3],
                    )
                    for e in rea_edge:
                        g.add_edge(
                            e[0], e[1], color=e[2], font_color="black",
                        )
            for pro in r.products.get("assembly"):
                tmp_pro = re.sub("_([cep])$", "", pro.id)
                if pro.id in path_met:
                    colname_pro = define_node_features(
                        color_brewer, "metabolites", pro, n_letter,
                    )
                    pro_edge = define_edge_features(
                        color_brewer,
                        "single_path_r",
                        "metabolites",
                        pro,
                        colname_pro[1],
                        r,
                        colname_r[1],
                        "products",
                        n_letter,
                    )
                    g.add_node(
                        colname_pro[0],
                        shape="o",
                        label=colname_pro[1],
                        color=colname_pro[2],
                        title=colname_pro[3],
                    )
                    for e in pro_edge:
                        g.add_edge(
                            e[0], e[1], color=e[2], font_color="black",
                        )
                elif additional_met:
                    if tmp_pro not in met_not_int.keys():
                        colname_pro = define_node_features(
                            color_brewer, "metConnect", pro, n_letter,
                        )
                        pro_edge = define_edge_features(
                            color_brewer,
                            "single_path_r",
                            "metabolites",
                            pro,
                            colname_pro[1],
                            r,
                            colname_r[1],
                            "products",
                            n_letter,
                        )
                    else:
                        colname_pro = define_node_features(
                            color_brewer,
                            "metConnect",
                            pro,
                            n_letter,
                            node_id=f"{tmp_pro} {met_not_int[tmp_pro]}",
                            label=tmp_pro,
                        )
                        pro_edge = define_edge_features(
                            color_brewer,
                            "single_path_r",
                            "metabolites",
                            pro,
                            f"{tmp_pro} {met_not_int[tmp_pro]}",
                            r,
                            colname_r[1],
                            "products",
                            n_letter,
                        )
                        met_not_int[tmp_pro] = met_not_int[tmp_pro] + 1
                    g.add_node(
                        colname_pro[0],
                        shape="o",
                        label=colname_pro[1],
                        color=colname_pro[2],
                        title=colname_pro[3],
                    )
                    for e in pro_edge:
                        g.add_edge(
                            e[0], e[1], color=e[2], font_color="black",
                        )
            if genes:
                g_colname = define_node_features(
                    color_brewer,
                    "genes",
                    r,
                    n_letter,
                    gene=True,
                    and_as_solid=and_as_solid,
                )
                g.add_node(
                    g_colname[0],
                    shape="box",
                    label=g_colname[1],
                    color=g_colname[2],
                    title=g_colname[3],
                )
                g.add_edge(
                    r_id, g_colname[0], color=g_colname[2], font_color="black",
                )
        else:
            not_f_col = color_brewer["notFound"][
                math.floor(len(supermodel.sources) / 2)
            ]
            g.add_node(
                r_id,
                label=r_id,
                shape="box",
                color=not_f_col,
                title="Confirmed by:\nNone",
            )
            for pair in pathway.get(r_id):
                if pair[0] in supermodel.metabolites.assembly.keys():
                    colname_rea = define_node_features(
                        color_brewer,
                        "metabolites",
                        supermodel.metabolites.assembly[pair[0]],
                        n_letter,
                    )
                else:
                    colname_rea = [pair[0], pair[0], not_f_col, "Confirmed by:\nNone"]
                if pair[1] in supermodel.metabolites.assembly.keys():
                    colname_pro = define_node_features(
                        color_brewer,
                        "metabolites",
                        supermodel.metabolites.assembly[pair[1]],
                        n_letter,
                    )
                else:
                    colname_pro = [pair[1], pair[1], not_f_col, "Confirmed by:\nNone"]
                g.add_node(
                    colname_rea[0],
                    label=colname_rea[1],
                    shape="o",
                    color=colname_rea[2],
                    title=colname_rea[3],
                )
                g.add_node(
                    colname_pro[0],
                    label=colname_pro[1],
                    shape="o",
                    color=colname_pro[2],
                    title=colname_pro[3],
                )
                g.add_edge(
                    colname_rea[1], r_id, color=not_f_col, font_color="black",
                )
                g.add_edge(
                    r_id, colname_pro[1], color=not_f_col, font_color="black",
                )
    pyvis_graph = get_pyvis_from_nx(g, directed, size, wid, hei)
    pyvis_graph.write_html(output_name, notebook=False)
    return g


def draw_one_synt_path(
    supermodel: SuperModel,
    path: list,
    medium: list,
    met_to_synt: list,
    output_name: str,
    draw_met_not_int=False,
    genes=True,
    and_as_solid=False,
    directed=True,
    met_not_int=None,
    n_letter=None,
    wid=1920,
    hei=1080,
    size=25,
):
    if not output_name.endswith(".html"):
        raise ValueError(
            "Output file for the plot is of wrong format. Please use html file name."
        )
    if met_not_int is None:
        met_not_int = deepcopy(MET_NOT_INT_GLOBAL)
    color_brewer = get_color_palette(len(supermodel.sources))
    if n_letter is None:
        n_letter = supermodel.get_short_name_len()
    g = nx.DiGraph()
    for r_id in path:
        r = supermodel.reactions.assembly.get(r_id)
        colname_r = define_node_features(color_brewer, "single_path_r", r, n_letter,)
        g.add_node(
            colname_r[0],
            label=colname_r[1],
            shape="box",
            color=colname_r[2],
            title=colname_r[3],
        )
        for rea in r.reactants.get("assembly"):
            tmp_rea = re.sub("_([cep])$", "", rea.id)
            if (tmp_rea in met_not_int.keys()) and not draw_met_not_int:
                continue
            if (tmp_rea in met_not_int.keys()) and draw_met_not_int:
                if rea.id in met_to_synt:
                    m_pallitra = "interest"
                else:
                    m_pallitra = "metConnect"
                colname_rea = define_node_features(
                    color_brewer,
                    m_pallitra,
                    rea,
                    n_letter,
                    node_id=f"{tmp_rea} {met_not_int[tmp_rea]}",
                    label=tmp_rea,
                )
                rea_edge = define_edge_features(
                    color_brewer,
                    "single_path_r",
                    "metabolites",
                    rea,
                    f"{tmp_rea} {met_not_int[tmp_rea]}",
                    r,
                    colname_r[1],
                    "reactants",
                    n_letter,
                )
                met_not_int[tmp_rea] = met_not_int[tmp_rea] + 1
            else:
                if rea.id in medium:
                    m_pallitra = "metConnect"
                elif rea.id in met_to_synt:
                    m_pallitra = "interest"
                else:
                    m_pallitra = "metabolites"
                colname_rea = define_node_features(
                    color_brewer, m_pallitra, rea, n_letter,
                )
                rea_edge = define_edge_features(
                    color_brewer,
                    "single_path_r",
                    "metabolites",
                    rea,
                    colname_rea[1],
                    r,
                    colname_r[1],
                    "reactants",
                    n_letter,
                )
            g.add_node(
                colname_rea[0],
                shape="o",
                label=colname_rea[1],
                color=colname_rea[2],
                title=colname_rea[3],
            )
            for e in rea_edge:
                g.add_edge(
                    e[0], e[1], color=e[2], font_color="black",
                )
        for pro in r.products.get("assembly"):
            tmp_pro = re.sub("_([cep])$", "", pro.id)
            if (tmp_pro in met_not_int.keys()) and not draw_met_not_int:
                continue
            if (tmp_pro in met_not_int.keys()) and draw_met_not_int:
                if pro.id in met_to_synt:
                    m_pallitra = "interest"
                else:
                    m_pallitra = "metConnect"
                colname_pro = define_node_features(
                    color_brewer,
                    m_pallitra,
                    pro,
                    n_letter,
                    node_id=f"{tmp_pro} {met_not_int[tmp_pro]}",
                    label=tmp_pro,
                )
                pro_edge = define_edge_features(
                    color_brewer,
                    "single_path_r",
                    "metabolites",
                    pro,
                    f"{tmp_pro} {met_not_int[tmp_pro]}",
                    r,
                    colname_r[1],
                    "products",
                    n_letter,
                )
                met_not_int[tmp_pro] = met_not_int[tmp_pro] + 1
            else:
                if pro.id in medium:
                    m_pallitra = "metConnect"
                elif pro.id in met_to_synt:
                    m_pallitra = "interest"
                else:
                    m_pallitra = "metabolites"
                colname_pro = define_node_features(
                    color_brewer, m_pallitra, pro, n_letter,
                )
                pro_edge = define_edge_features(
                    color_brewer,
                    "single_path_r",
                    "metabolites",
                    pro,
                    colname_pro[1],
                    r,
                    colname_r[1],
                    "products",
                    n_letter,
                )
            g.add_node(
                colname_pro[0],
                shape="o",
                label=colname_pro[1],
                color=colname_pro[2],
                title=colname_pro[3],
            )
            for e in pro_edge:
                g.add_edge(
                    e[0], e[1], color=e[2], font_color="black",
                )
        if genes:
            g_colname = define_node_features(
                color_brewer,
                "genes",
                r,
                n_letter,
                gene=True,
                and_as_solid=and_as_solid,
            )
            g.add_node(
                g_colname[0],
                shape="box",
                label=g_colname[1],
                color=g_colname[2],
                title=g_colname[3],
            )
            g.add_edge(
                r_id, g_colname[0], color=g_colname[2], font_color="black",
            )
    pyvis_graph = get_pyvis_from_nx(g, directed, size, wid, hei)
    pyvis_graph.write_html(output_name, notebook=False)
    return g



from importlib.resources import files

from cobra.core.model import Model

from gemsembler import GatheredModels
from gemsembler.conversion import (
    ConvAgora,
    ConvBase,
    ConvCarveme,
    ConvGapseq,
    ConvModelseed,
)
from gemsembler.data import BU


class TestGathering:
    def test_gathered_models(self):
        # Test GatheredModels init method
        g = GatheredModels()

        # Check configuration in GatheredModels object
        conf = g.get_conf()

        assert "agora" in conf
        assert conf.get("agora").get("remove_b") == False
        assert conf.get("agora").get("db_name") == "weird_bigg"
        assert conf.get("agora").get("wo_periplasmic") == True
        assert type(conf.get("agora").get("conv_strategy")) is ConvAgora

        assert "carveme" in conf
        assert conf.get("carveme").get("remove_b") == False
        assert conf.get("carveme").get("db_name") == "bigg"
        assert conf.get("carveme").get("wo_periplasmic") == False
        assert type(conf.get("carveme").get("conv_strategy")) is ConvCarveme

        assert "gapseq" in conf
        assert conf.get("gapseq").get("remove_b") == False
        assert conf.get("gapseq").get("db_name") == "modelseed"
        assert conf.get("gapseq").get("wo_periplasmic") == False
        assert type(conf.get("gapseq").get("conv_strategy")) is ConvGapseq

        assert "modelseed" in conf
        assert conf.get("modelseed").get("remove_b") == True
        assert conf.get("modelseed").get("db_name") == "modelseed"
        assert conf.get("modelseed").get("wo_periplasmic") == True
        assert type(conf.get("modelseed").get("conv_strategy")) is ConvModelseed

        # Check models in GatheredModels object
        assert len(g) == 0

        g.add_model("test_carveme", files(BU) / "BU_carveme_hom.xml.gz", "carveme")
        assert "test_carveme" in g
        model_attrs = g.get_model_attrs("test_carveme")
        assert model_attrs["path_to_model"] == files(BU) / "BU_carveme_hom.xml.gz"
        assert model_attrs["model_type"] == "carveme"
        model = model_attrs["preprocess_model"]
        assert type(model) is Model
        assert len(model.metabolites) == 1244
        assert len(model.reactions) == 1849

        g.add_model("test_agora", files(BU) / "BU_agora.xml.gz", "agora")
        assert "test_agora" in g
        model_attrs = g.get_model_attrs("test_agora")
        assert model_attrs["path_to_model"] == files(BU) / "BU_agora.xml.gz"
        assert model_attrs["model_type"] == "agora"
        model = model_attrs["preprocess_model"]
        assert type(model) is Model
        assert len(model.metabolites) == 1500
        assert len(model.reactions) == 2418

        g.add_model("test_modelseed", files(BU) / "BU_modelSEED.sbml.gz", "modelseed")
        assert "test_modelseed" in g
        model_attrs = g.get_model_attrs("test_modelseed")
        assert model_attrs["path_to_model"] == files(BU) / "BU_modelSEED.sbml.gz"
        assert model_attrs["model_type"] == "modelseed"
        model = model_attrs["preprocess_model"]
        assert type(model) is Model
        assert len(model.metabolites) == 1271
        assert len(model.reactions) == 1161

        g.add_model("test_gapseq", files(BU) / "BU_gapseq.xml.gz", "gapseq")
        assert "test_gapseq" in g
        model_attrs = g.get_model_attrs("test_gapseq")
        assert model_attrs["path_to_model"] == files(BU) / "BU_gapseq.xml.gz"
        assert model_attrs["model_type"] == "gapseq"
        model = model_attrs["preprocess_model"]
        assert type(model) is Model
        assert len(model.metabolites) == 1520
        assert len(model.reactions) == 1891

        assert g.same_db_models["bigg"] == {"test_carveme": "carveme"}
        assert g.same_db_models["weird_bigg"] == {"test_agora": "agora"}
        assert g.same_db_models["modelseed"] == {
            "test_modelseed": "modelseed",
            "test_gapseq": "gapseq",
        }
        # TODO: finish the tests for gathered models
        g.run()



from gemsembler import GatheredModels, bu_example
from gemsembler.selection import run_selection


class TestSelection:
    def test_first_stage(self):
        g = GatheredModels()
        for model in bu_example:
            g.add_model(**model)

        sel_mbs = run_selection(g.same_db_models, g.converted_metabolites, "highest")

        # Check the number of metabolites is consistant from `run_selection`
        assert len(sel_mbs["carveme_BU"]) == 1244
        assert len(sel_mbs["agora_BU"]) == 1500
        assert len(sel_mbs["gapseq_BU"]) == 1520
        assert len(sel_mbs["modelseed_BU"]) == 1271

        # Check individual metabolite from carveme model
        sel_mb = sel_mbs["carveme_BU"]["10fthf_c"]
        assert sel_mb.compartments == ["c"]
        assert sel_mb.consistent == "Yes"
        assert sel_mb.highest_consistent == ["10fthf_c"]
        assert sel_mb.from_one_id == True
        assert sel_mb.to_one_id == True
        assert sel_mb.from_many_other_ids == []
        assert sel_mb.in_other_models == {}

        # Check individual metabolite from agora model
        sel_mb = sel_mbs["agora_BU"]["gam1p[c]"]
        assert sel_mb.compartments == ["c"]
        assert sel_mb.consistent == "Yes"
        assert sel_mb.highest_consistent == ["gam1p_c"]
        assert sel_mb.from_one_id == True
        assert sel_mb.to_one_id == True
        assert sel_mb.from_many_other_ids == []
        assert sel_mb.in_other_models == {}

        # Check individual metabolite from modelseed model
        sel_mb = sel_mbs["modelseed_BU"]["cpd02902_c0"]
        assert sel_mb.compartments == ["c"]
        assert sel_mb.consistent == "Not converted"
        assert sel_mb.highest_consistent == []
        assert sel_mb.from_one_id == False
        assert sel_mb.to_one_id == None
        assert len(sel_mb.from_many_other_ids) == 357
        assert sel_mb.in_other_models == {}

        # Check individual metabolite from gapseq model
        sel_mb = sel_mbs["gapseq_BU"]["cpd11477_c0"]
        assert sel_mb.compartments == ["c"]
        assert sel_mb.consistent == "Yes"
        assert sel_mb.highest_consistent == ["tpalm2eACP_c"]
        assert sel_mb.from_one_id == True
        assert sel_mb.to_one_id == True
        assert sel_mb.from_many_other_ids == []
        assert sel_mb.in_other_models == {"modelseed_BU": [True, True]}



import warnings
from copy import deepcopy
from pathlib import Path
from pprint import pprint

from cobra import Metabolite, Model, Reaction
from cobra.io import validate_sbml_model, write_sbml_model

from .creation import NewElement, SuperModel


def gapfill_transport_r(cobra_model: Model, supermodel: SuperModel):
    for exchange in cobra_model.exchanges:
        met_e = list(exchange.reactants)[0].id
        met_c = list(exchange.reactants)[0].id[:-1] + "c"
        cobra_transport = False
        for r in cobra_model.metabolites.get_by_id(met_e).reactions:
            r_react = [m.id for m in r.reactants]
            r_pro = [m.id for m in r.products]
            if ((met_e in r_react) & (met_c in r_pro)) | (
                (met_e in r_pro) & (met_c in r_react)
            ):
                cobra_transport = True
        if not cobra_transport:
            if met_c in supermodel.metabolites.assembly.keys():
                transport_r = []
                for r_super in supermodel.metabolites.assembly.get(met_e).reactions.get(
                    "assembly"
                ):
                    rs_react = [m.id for m in r_super.reactants.get("assembly")]
                    rs_pro = [m.id for m in r_super.products.get("assembly")]
                    if ((met_e in rs_react) & (met_c in rs_pro)) | (
                        (met_e in rs_pro) & (met_c in rs_react)
                    ):
                        transport_r.append(r_super)
                for tr in transport_r:
                    tr_r = Reaction(tr.id)
                    if tr.name:
                        tr_r.name = tr.name
                    else:
                        tr_r.name = ""
                    out_subsystem = ""
                    for source in tr.in_models["models_list"]:
                        out_subsystem = (
                            out_subsystem
                            + "#"
                            + source
                            + "#"
                            + tr.subsystem.get(source)[0]
                        )
                    tr_r.subsystem = out_subsystem
                    tr_r.lower_bound = tr.lower_bound.get("assembly")[0]
                    tr_r.upper_bound = tr.upper_bound.get("assembly")[0]
                    for met, k in tr.metabolites.get("assembly").items():
                        tr_met = Metabolite(
                            met.id, name=met.name, compartment=met.id[-1]
                        )
                        tr_r.add_metabolites({tr_met: k})
                    cobra_model.add_reactions([tr_r])


def get_model_of_interest(
    supermodel: SuperModel,
    interest_level: str,
    output_name=None,
    gene_interest_level=None,
    biomass_interest_level=None,
    simple_biomass_products=True,
    extend_zero_bounds=True,
    gapfill_transport=True,
    reactions_include: [NewElement] = None,
    reactions_exclude: [NewElement] = None,
):
    """ Creating COBRA model from supermodel based on specific level of interest for example core or union.
    Additionaly, some reactions"""
    if not gene_interest_level:
        gene_interest_level = interest_level
    if not biomass_interest_level:
        biomass_interest_level = interest_level
    outmodel = Model(interest_level)
    if interest_level in supermodel.sources + ["assembly"]:
        in_reactions = getattr(supermodel.reactions, interest_level).values()
    elif interest_level in supermodel.reactions.comparison.keys():
        in_reactions = supermodel.reactions.comparison[interest_level].values()
    else:
        raise ValueError(
            f"Interest level {interest_level} is not determined yet. "
            f"Please run corresponding supermodel comparison first."
        )
    if reactions_include:
        in_reactions = list(set(in_reactions) | set(reactions_include))
    else:
        reactions_include = []
    if reactions_exclude:
        in_reactions = list(set(in_reactions) - set(reactions_exclude))
    if supermodel.reactions.assembly.get("Biomass") in in_reactions:
        in_reactions = list(
            set(list(in_reactions)) - {supermodel.reactions.assembly.get("Biomass")}
        )
    for r in in_reactions:
        if r in reactions_include:
            interest_level_r = "assembly"
            gene_interest_level_r = "assembly"
        else:
            interest_level_r = interest_level
            gene_interest_level_r = gene_interest_level
        if interest_level_r in supermodel.sources + ["assembly"]:
            r_upper_bound = r.upper_bound
            r_lower_bound = r.lower_bound
            r_metabolites = r.metabolites
        else:
            r_upper_bound = r.upper_bound["comparison"]
            r_lower_bound = r.lower_bound["comparison"]
            r_metabolites = r.metabolites["comparison"]
        if gene_interest_level_r in supermodel.sources + ["assembly"]:
            r_gene_reaction_rule = r.gene_reaction_rule
        else:
            r_gene_reaction_rule = r.gene_reaction_rule["comparison"]
        out_reaction = Reaction(r.id)
        if r.name:
            out_reaction.name = r.name
        else:
            out_reaction.name = ""
        out_subsystem = ""
        for source in r.in_models["models_list"]:
            out_subsystem = (
                out_subsystem + "#" + source + "#" + r.subsystem.get(source)[0]
            )
        out_reaction.subsystem = out_subsystem
        if extend_zero_bounds and (
            r_upper_bound.get(interest_level_r)[0]
            - r_lower_bound.get(interest_level_r)[0]
            == 0
        ):
            out_reaction.lower_bound = -1000.0
            out_reaction.upper_bound = 1000.0
        else:
            out_reaction.lower_bound = r_lower_bound.get(interest_level_r)[0]
            out_reaction.upper_bound = r_upper_bound.get(interest_level_r)[0]
        for met, k in r_metabolites.get(interest_level_r).items():
            out_met = Metabolite(
                met.id, name=met.name, compartment=met.compartments["assembly"][0]
            )
            out_reaction.add_metabolites({out_met: k})
        if r_gene_reaction_rule.get(gene_interest_level_r):
            out_reaction.gene_reaction_rule = r_gene_reaction_rule.get(
                gene_interest_level_r
            )[0]
        else:
            out_reaction.gene_reaction_rule = ""
        outmodel.add_reactions([out_reaction])
    biomass_reaction = Reaction("Biomass")
    if biomass_interest_level in supermodel.sources + ["assembly"]:
        biomass_reaction.upper_bound = supermodel.reactions.assembly[
            "Biomass"
        ].upper_bound.get(biomass_interest_level)[0]
        biomass_reaction.lower_bound = supermodel.reactions.assembly[
            "Biomass"
        ].lower_bound.get(biomass_interest_level)[0]
        bio_r_metabolites = supermodel.reactions.assembly["Biomass"].metabolites
    else:
        biomass_reaction.upper_bound = (
            supermodel.reactions.assembly["Biomass"]
            .upper_bound["comparison"]
            .get(biomass_interest_level)[0]
        )
        biomass_reaction.lower_bound = (
            supermodel.reactions.assembly["Biomass"]
            .lower_bound["comparison"]
            .get(biomass_interest_level)[0]
        )
        bio_r_metabolites = supermodel.reactions.assembly["Biomass"].metabolites[
            "comparison"
        ]
    if simple_biomass_products:
        biomass_prod_id = ["adp_c", "h_c", "pi_c", "ppi_c"]
        for met, k in bio_r_metabolites.get(biomass_interest_level).items():
            if met.id in biomass_prod_id:
                biomass_prod_id.remove(met.id)
                bio_met = Metabolite(
                    met.id, name=met.name, compartment=met.compartments["assembly"][0]
                )
                biomass_reaction.add_metabolites({bio_met: k})
            if k < 0:
                bio_met = Metabolite(
                    met.id, name=met.name, compartment=met.compartments["assembly"][0]
                )
                biomass_reaction.add_metabolites({bio_met: k})
        if biomass_prod_id:
            warnings.warn(
                f"Some expected biomass products are not found: {' '.join(biomass_prod_id)}"
            )
    else:
        for met, k in bio_r_metabolites.get(biomass_interest_level).items():
            bio_met = Metabolite(
                met.id, name=met.name, compartment=met.compartments["assembly"][0]
            )
            biomass_reaction.add_metabolites({bio_met: k})
    outmodel.add_reactions([biomass_reaction])
    outmodel.objective = "Biomass"
    if gapfill_transport:
        gapfill_transport_r(outmodel, supermodel)
    if output_name is not None:
        write_sbml_model(outmodel, output_name)
        report = validate_sbml_model(filename=output_name)
        pprint(report)
    return outmodel


def get_models_with_all_confidence_levels(
    supermodel: SuperModel,
    output_dir=None,
    gene_interest_level=None,
    biomass_interest_level=None,
    simple_biomass_products=True,
    extend_zero_bounds=True,
    gapfill_transport=True,
    reactions_include: [NewElement] = None,
    reactions_exclude: [NewElement] = None,
):
    output_models = {}
    confidence_levels = deepcopy(supermodel.sources)
    for i in range(len(supermodel.sources), 1, -1):
        confidence_levels.append("core" + str(i))
    confidence_levels.append("assembly")
    for level in confidence_levels:
        if output_dir is not None:
            output_dir_lev = Path(output_dir) / (level + ".xml")
        else:
            output_dir_lev = None
        output_models.update(
            {
                level: get_model_of_interest(
                    supermodel,
                    level,
                    output_name=output_dir_lev,
                    gene_interest_level=gene_interest_level,
                    biomass_interest_level=biomass_interest_level,
                    simple_biomass_products=simple_biomass_products,
                    extend_zero_bounds=extend_zero_bounds,
                    gapfill_transport=gapfill_transport,
                    reactions_include=reactions_include,
                    reactions_exclude=reactions_exclude,
                )
            }
        )
    return output_models



import itertools
from collections import defaultdict
from copy import deepcopy
from itertools import combinations

import cobra

from .selection import Selected


class StructuralR(object):
    def __init__(self, bigg_structural: dict, selected: Selected):
        comment = list(bigg_structural.values())[0]
        if list(bigg_structural.keys())[0] == "NOT_found" or comment.startswith(
            "Potentially_found"
        ):
            self.structural = []
        else:
            self.structural = list(bigg_structural.keys())
        if len(bigg_structural) == 1 and comment.startswith(
            "Found_via_adding_periplasmic_compartment"
        ):
            self.comment = "Found_via_adding_periplasmic_compartment"
            self.suggestions = {"orig_m": [], "b_m": [], "p_m": []}
            if comment.split("-")[1].split(" ")[0] != "":
                self.suggestions["orig_m"] = self.suggestions["orig_m"] + comment.split(
                    "-"
                )[1].split(" ")
                self.suggestions["b_m"] = self.suggestions["b_m"] + comment.split("-")[
                    2
                ].split(" ")
                self.suggestions["p_m"] = self.suggestions["p_m"] + comment.split("-")[
                    3
                ].split(" ")
            if comment.split("-")[4].split(" ")[0] != "":
                self.suggestions["orig_m"] = self.suggestions["orig_m"] + comment.split(
                    "-"
                )[4].split(" ")
                self.suggestions["b_m"] = self.suggestions["b_m"] + comment.split("-")[
                    5
                ].split(" ")
                self.suggestions["p_m"] = self.suggestions["p_m"] + comment.split("-")[
                    6
                ].split(" ")
            self.compartments = list(bigg_structural.values())[0].split("-")[7].split()
        else:
            self.compartments = selected.compartments
            if len(bigg_structural) == 1 and comment.startswith(
                "Found_via_one_to_many_metabolites"
            ):
                self.comment = "Found_via_one_to_many_metabolites"
                self.suggestions = {
                    "orig_m": comment.split("-")[1].split(" "),
                    "b_m": comment.split("-")[2].split(" "),
                }
            elif len(bigg_structural) == 1 and comment.startswith(
                "Potentially_found_via_many_to_one_metabolites"
            ):
                self.comment = "Potentially_found_via_many_to_one_metabolites"
                self.suggestions = {
                    "reaction": list(bigg_structural.keys())[0],
                    "orig_m": [],
                    "b_m": [],
                }
                if comment.split("-")[1].split(" ")[0] != "":
                    self.suggestions["orig_m"] = self.suggestions[
                        "orig_m"
                    ] + comment.split("-")[1].split(" ")
                    self.suggestions["b_m"] = self.suggestions["b_m"] + comment.split(
                        "-"
                    )[2].split(" ")
                if comment.split("-")[3].split(" ")[0] != "":
                    self.suggestions["orig_m"] = self.suggestions[
                        "orig_m"
                    ] + comment.split("-")[3].split(" ")
                    self.suggestions["b_m"] = self.suggestions["b_m"] + comment.split(
                        "-"
                    )[4].split(" ")
            elif len(bigg_structural) == 1 and comment.startswith(
                (
                    "Not_found_via_many_to_one_metabolites",
                    "Potentially_found_but_no_confidence",
                    "Not_all_metabolites_are_converted_but_important_for_many_original",
                )
            ):
                if comment.startswith("Not_found_via_many_to_one_metabolites"):
                    list_name = "not_fit"
                else:
                    list_name = "no_data"
                self.comment = comment.split("-")[0]
                self.suggestions = {
                    "reaction": list(bigg_structural.keys())[0],
                    "orig_m": [],
                    "b_m": [],
                }
                if comment.split("-")[1].split(" ")[0] != "":
                    self.suggestions["orig_m"] = self.suggestions[
                        "orig_m"
                    ] + comment.split("-")[1].split(" ")
                    self.suggestions["b_m"] = self.suggestions["b_m"] + [
                        list_name
                    ] * len(comment.split("-")[1].split(" "))
                if comment.split("-")[2].split(" ")[0] != "":
                    self.suggestions["orig_m"] = self.suggestions[
                        "orig_m"
                    ] + comment.split("-")[2].split(" ")
                    self.suggestions["b_m"] = self.suggestions["b_m"] + [
                        list_name
                    ] * len(comment.split("-")[2].split(" "))
            else:
                self.suggestions = None
                if len(bigg_structural) == 1:
                    self.comment = comment
                else:
                    self.comment = f"Several result ids and comments: {' -NEW_COMMENT- '.join(list(bigg_structural.values()))}"


class StructuralM(object):
    def __init__(self, ids: list, comment: str, compartments: list):
        self.structural = ids
        self.compartments = compartments
        self.comment = comment


def getReaction(bigg_met1, bigg_met2, bigg_network_r, comment):
    """ Find reaction id from reaction's metabolites """

    bigg_met1_str = " ".join(sorted(bigg_met1))
    bigg_met2_str = " ".join(sorted(bigg_met2))
    bigg_equation = "<->".join(sorted([bigg_met1_str, bigg_met2_str]))
    if bigg_equation in list(bigg_network_r.keys()):
        return {bigg_network_r[bigg_equation]: comment}
    else:
        return {}


def Hydrogens(
    bigg_met1, bigg_met2, compart1, compart2, bigg_network_r, addit_comment=""
):
    """ Trying to convert reaction after add/removing H from/to reactants/products """
    bigg_r = {}
    if addit_comment != "":
        addit_comment = addit_comment + "-"
    for c1 in list(set(compart1)):
        bigg_met1_mod = deepcopy(bigg_met1)
        if "h_" + c1 in bigg_met1_mod:
            bigg_met1_mod.remove("h_" + c1)
            tmp_bigg_r = getReaction(
                bigg_met1_mod,
                bigg_met2,
                bigg_network_r,
                addit_comment + "Found_via_removing_H_from_reactants",
            )
            if tmp_bigg_r:
                bigg_r.update(tmp_bigg_r)
            else:
                for cc2 in list(set(compart2)):
                    bigg_met2_h = deepcopy(bigg_met2)
                    bigg_met2_h.append("h_" + cc2)
                    tmp_bigg_r = getReaction(
                        bigg_met1_mod,
                        bigg_met2_h,
                        bigg_network_r,
                        addit_comment
                        + "Found_via_removing_H_from_reactants_adding_H_to_products",
                    )
                    if tmp_bigg_r:
                        bigg_r.update(tmp_bigg_r)
        else:
            bigg_met1_mod.append("h_" + c1)
            tmp_bigg_r = getReaction(
                bigg_met1_mod,
                bigg_met2,
                bigg_network_r,
                addit_comment + "Found_via_adding_H_to_reactants",
            )
            if tmp_bigg_r:
                bigg_r.update(tmp_bigg_r)
    for c2 in list(set(compart2)):
        bigg_met2_mod = deepcopy(bigg_met2)
        if "h_" + c2 in bigg_met2_mod:
            bigg_met2_mod.remove("h_" + c2)
            tmp_bigg_r = getReaction(
                bigg_met1,
                bigg_met2_mod,
                bigg_network_r,
                addit_comment + "Found_via_removing_H_from_products",
            )
            if tmp_bigg_r:
                bigg_r.update(tmp_bigg_r)
            else:
                for cc1 in list(set(compart1)):
                    bigg_met1_h = deepcopy(bigg_met1)
                    bigg_met1_h.append("h_" + cc1)
                    tmp_bigg_r = getReaction(
                        bigg_met1_h,
                        bigg_met2_mod,
                        bigg_network_r,
                        addit_comment
                        + "Found_via_adding_H_from_reactants_removing_H_to_products",
                    )
                    if tmp_bigg_r:
                        bigg_r.update(tmp_bigg_r)
        else:
            bigg_met2_mod.append("h_" + c2)
            tmp_bigg_r = getReaction(
                bigg_met1,
                bigg_met2_mod,
                bigg_network_r,
                addit_comment + "Found_via_adding_H_to_products",
            )
            if tmp_bigg_r:
                bigg_r.update(tmp_bigg_r)
    return bigg_r


def Periplasmic(bigg_met1, bigg_met2, bigg_network_r):
    """ Trying to convert reaction after replacing compartments to periplasmic for all possible set of metabolites """
    bigg_r = {}
    bigg1_comb = []
    for r1 in range(len(bigg_met1) + 1):
        bigg1_comb = bigg1_comb + list(combinations(list(bigg_met1.keys()), r1))
    bigg2_comb = []
    for r2 in range(len(bigg_met2) + 1):
        bigg2_comb = bigg2_comb + list(combinations(list(bigg_met2.keys()), r2))
    for comb1 in bigg1_comb:
        for comb2 in bigg2_comb:
            bigg1 = list(set(bigg_met1) - set(comb1))
            c1 = list(set([b1[-1] for b1 in bigg1]))
            orig1 = [v for k, v in bigg_met1.items() if k in comb1]
            if comb1:
                bigg1_p = [m1[:-1] + "p" for m1 in comb1]
                c1.append("p")
            else:
                bigg1_p = []
            bigg2 = list(set(bigg_met2) - set(comb2))
            c2 = list(set([b2[-1] for b2 in bigg2]))
            orig2 = [vv for kk, vv in bigg_met2.items() if kk in comb2]
            if comb2:
                bigg2_p = [m2[:-1] + "p" for m2 in comb2]
                c2.append("p")
            else:
                bigg2_p = []
            tmp_bigg_r = getReaction(
                bigg1 + bigg1_p,
                bigg2 + bigg2_p,
                bigg_network_r,
                f"Found_via_adding_periplasmic_compartment-{' '.join(orig1)}-{' '.join(comb1)}-{' '.join(bigg1_p)}"
                f"-{' '.join(orig2)}-{' '.join(comb2)}-{' '.join(bigg2_p)}-{' '.join(list(set(c1 + c2)))}",
            )
            if tmp_bigg_r:
                bigg_r.update(tmp_bigg_r)
    return bigg_r


def SeveralMetabolites(
    bigg_met1,
    bigg_met2,
    met1_to_many,
    met2_to_many,
    compart1,
    compart2,
    bigg_network_r,
):
    """ Trying to convert reaction with several variants coming from 1-n metabolites by variants from 1-n metabolites to 1-1 converted metabolites """
    bigg_r = {}
    bigg_r_h = {}
    bigg_to_many_variants1 = []
    orig_to_many_variants1 = []
    if met1_to_many != {}:
        for key1, val1 in met1_to_many.items():
            orig_to_many_variants1.append(key1)
            bigg_to_many_variants1.append([met_variant1 for met_variant1 in val1])
        bigg_to_many_variants1 = list(itertools.product(*bigg_to_many_variants1))
    bigg_to_many_variants2 = []
    orig_to_many_variants2 = []
    if met2_to_many != {}:
        for key2, val2 in met2_to_many.items():
            orig_to_many_variants2.append(key2)
            bigg_to_many_variants2.append([met_variant2 for met_variant2 in val2])
        bigg_to_many_variants2 = list(itertools.product(*bigg_to_many_variants2))
    if bigg_to_many_variants1 and bigg_to_many_variants2:
        for variant1 in bigg_to_many_variants1:
            for variant2 in bigg_to_many_variants2:
                bigg_met1_mod = list(bigg_met1.keys()) + list(variant1)
                bigg_met2_mod = list(bigg_met2.keys()) + list(variant2)
                tmp_bigg_r = getReaction(
                    bigg_met1_mod,
                    bigg_met2_mod,
                    bigg_network_r,
                    f"Found_via_one_to_many_metabolites-{' '.join(orig_to_many_variants1)}-{' '.join(variant1)}-{' '.join(orig_to_many_variants2)}-{' '.join(variant2)}",
                )
                tmp_bigg_r_h = Hydrogens(
                    bigg_met1_mod,
                    bigg_met2_mod,
                    compart1,
                    compart2,
                    bigg_network_r,
                    f"Found_via_one_to_many_metabolites-{' '.join(orig_to_many_variants1)}-{' '.join(variant1)}-{' '.join(orig_to_many_variants2)}-{' '.join(variant2)}",
                )

                if tmp_bigg_r:
                    bigg_r.update(tmp_bigg_r)
                if tmp_bigg_r_h:
                    bigg_r_h.update(tmp_bigg_r_h)
    elif bigg_to_many_variants1 and (not bigg_to_many_variants2):
        for variant1 in bigg_to_many_variants1:
            bigg_met1_mod = list(bigg_met1.keys()) + list(variant1)
            tmp_bigg_r = getReaction(
                bigg_met1_mod,
                list(bigg_met2.keys()),
                bigg_network_r,
                f"Found_via_one_to_many_metabolites-{' '.join(orig_to_many_variants1)}-{' '.join(variant1)}",
            )
            tmp_bigg_r_h = Hydrogens(
                bigg_met1_mod,
                list(bigg_met2.keys()),
                compart1,
                compart2,
                bigg_network_r,
                f"Found_via_one_to_many_metabolites-{' '.join(orig_to_many_variants1)}-{' '.join(variant1)}",
            )
            if tmp_bigg_r:
                bigg_r.update(tmp_bigg_r)
            if tmp_bigg_r_h:
                bigg_r_h.update(tmp_bigg_r_h)
    elif bigg_to_many_variants2 and (not bigg_to_many_variants1):
        for variant2 in list(bigg_to_many_variants2):
            bigg_met2_mod = list(bigg_met2.keys()) + list(variant2)
            tmp_bigg_r = getReaction(
                list(bigg_met1.keys()),
                bigg_met2_mod,
                bigg_network_r,
                f"Found_via_one_to_many_metabolites-{' '.join(orig_to_many_variants2)}-{' '.join(variant2)}",
            )
            tmp_bigg_r_h = Hydrogens(
                list(bigg_met1.keys()),
                bigg_met2_mod,
                compart1,
                compart2,
                bigg_network_r,
                f"Found_via_one_to_many_metabolites-{' '.join(orig_to_many_variants2)}-{' '.join(variant2)}",
            )
            if tmp_bigg_r:
                bigg_r.update(tmp_bigg_r)
            if tmp_bigg_r_h:
                bigg_r_h.update(tmp_bigg_r_h)
    if bigg_r:
        return bigg_r
    else:
        return bigg_r_h


def convertReactionViaNetworkStructure(
    orig_met1: list,
    orig_met2: list,
    selected_met: dict,
    bigg_network_r: dict,
    do_priplasmic: bool,
):
    """ Converting one reaction with several strategies. 1) Try just via it's metabolites if all them are converted 1-1.
     2) If not successful try to add/remove H 3) If not successful try to consider periplasmic compartment 4) If all
     metabolites in the reaction are converted 1-1 or 1-n try different variants from this n options. Writing comments
     with information how it was converted and what happened to metabolites for further suggestions"""
    compart1 = []
    bigg_met1 = {}
    bigg_met1_to_many = {}
    bigg_met1_from_many = {}
    for met1 in orig_met1:
        compart1 = compart1 + selected_met[met1].compartments
        if (
            selected_met[met1].to_one_id == True
            and selected_met[met1].from_one_id == True
        ):
            bigg_met1.update({selected_met[met1].highest_consistent[0]: met1})
        elif (
            selected_met[met1].to_one_id == False
            and selected_met[met1].from_one_id == True
        ):
            bigg_met1_to_many.update({met1: selected_met[met1].highest_consistent})
        elif (
            selected_met[met1].to_one_id == True
            and selected_met[met1].from_one_id == False
        ):
            bigg_met1_from_many.update({met1: selected_met[met1].highest_consistent[0]})
    compart2 = []
    bigg_met2 = {}
    bigg_met2_to_many = {}
    bigg_met2_from_many = {}
    for met2 in orig_met2:
        compart2 = compart2 + selected_met[met2].compartments
        if (
            selected_met[met2].to_one_id == True
            and selected_met[met2].from_one_id == True
        ):
            bigg_met2.update({selected_met[met2].highest_consistent[0]: met2})
        elif (
            selected_met[met2].to_one_id == False
            and selected_met[met2].from_one_id == True
        ):
            bigg_met2_to_many.update({met2: selected_met[met2].highest_consistent})
        elif (
            selected_met[met2].to_one_id == True
            and selected_met[met2].from_one_id == False
        ):
            bigg_met2_from_many.update({met2: selected_met[met2].highest_consistent[0]})
    if (len(bigg_met1) == len(orig_met1)) & (len(bigg_met2) == len(orig_met2)):
        bigg_r = getReaction(
            list(bigg_met1.keys()),
            list(bigg_met2.keys()),
            bigg_network_r,
            "Found_via_pure_reaction_equation",
        )
        if not bigg_r:
            bigg_r = Hydrogens(
                list(bigg_met1.keys()),
                list(bigg_met2.keys()),
                compart1,
                compart2,
                bigg_network_r,
            )
            if not bigg_r:
                if do_priplasmic:
                    bigg_r = Periplasmic(bigg_met1, bigg_met2, bigg_network_r,)
                if not bigg_r:
                    bigg_r = {"NOT_found": "Not_found_in_BiGG_network"}
    elif (len(bigg_met1) + len(bigg_met1_to_many) == len(orig_met1)) & (
        len(bigg_met2) + len(bigg_met2_to_many) == len(orig_met2)
    ):
        bigg_r = SeveralMetabolites(
            bigg_met1,
            bigg_met2,
            bigg_met1_to_many,
            bigg_met2_to_many,
            compart1,
            compart2,
            bigg_network_r,
        )
        if not bigg_r:
            bigg_r = {"NOT_found": "Not_found_via_one_to_many_metabolites"}
    elif (len(bigg_met1) + len(bigg_met1_from_many) == len(orig_met1)) & (
        len(bigg_met2) + len(bigg_met2_from_many) == len(orig_met2)
    ):
        if len(bigg_met1) + len(bigg_met2) == 0:
            bigg_r = getReaction(
                list(bigg_met1.keys()) + list(bigg_met1_from_many.values()),
                list(bigg_met2.keys()) + list(bigg_met2_from_many.values()),
                bigg_network_r,
                f"Potentially_found_but_no_confidence-"
                f"{' '.join(list(bigg_met1_from_many.keys()))}-{' '.join(list(bigg_met2_from_many.keys()))}",
            )
        else:
            bigg_r = getReaction(
                list(bigg_met1.keys()) + list(bigg_met1_from_many.values()),
                list(bigg_met2.keys()) + list(bigg_met2_from_many.values()),
                bigg_network_r,
                f"Potentially_found_via_many_to_one_metabolites-"
                f"{' '.join(list(bigg_met1_from_many.keys()))}-{' '.join(list(bigg_met1_from_many.values()))}-"
                f"{' '.join(list(bigg_met2_from_many.keys()))}-{' '.join(list(bigg_met2_from_many.values()))}",
            )
        if not bigg_r:
            bigg_r = {
                "NOT_found": f"Not_found_via_many_to_one_metabolites-"
                f"{' '.join(list(bigg_met1_from_many.keys()))}-{' '.join(list(bigg_met2_from_many.keys()))}"
            }
    elif len(bigg_met1_from_many) + len(bigg_met2_from_many) > 0:
        bigg_r = {
            "NOT_found": f"Not_all_metabolites_are_converted_but_important_for_many_original-"
            f"{' '.join(list(bigg_met1_from_many.keys()))}-{' '.join(list(bigg_met2_from_many.keys()))}"
        }
    else:
        bigg_r = {"NOT_found": "Not_all_metabolites_for_reaction_are_converted"}
    return bigg_r


def runStructuralCheck(
    react_checked: dict,
    met_checked: dict,
    model: cobra.core.model.Model,
    bigg_network: dict,
):
    """ Checking reactions equations for models with no conversion need (with BiGG ids originally). Should be in BiGG
    database if not exchange or biomass reaction. """
    react_struct_checked = {}
    for id_r, sel in react_checked.items():
        bigg_met1 = [met1.id for met1 in model.reactions.get_by_id(id_r).reactants]
        bigg_met2 = [met2.id for met2 in model.reactions.get_by_id(id_r).products]
        bigg_met1_checked = []
        bigg_met2_checked = []
        for m1 in bigg_met1:
            if (
                met_checked[m1].to_one_id == True
                and met_checked[m1].from_one_id == True
            ):
                bigg_met1_checked.append(met_checked[m1].highest_consistent[0])
        for m2 in bigg_met2:
            if (
                met_checked[m2].to_one_id == True
                and met_checked[m2].from_one_id == True
            ):
                bigg_met2_checked.append(met_checked[m2].highest_consistent[0])
        if (len(bigg_met1) > 20) | (len(bigg_met2) > 20):
            react_struct_checked.update(
                {id_r: StructuralR({"Biomass": "growth_reaction"}, sel)}
            )
            continue
        bigg_r = getReaction(
            bigg_met1_checked,
            bigg_met2_checked,
            bigg_network,
            "checked_via_structural_reaction_equation",
        )
        if bigg_r:
            react_struct_checked.update({id_r: StructuralR(bigg_r, sel)})
        elif sel.highest_consistent:
            react_struct_checked.update(
                {
                    id_r: StructuralR(
                        {sel.highest_consistent[0]: "id_in_bigg_originally"}, sel
                    )
                }
            )
        elif (
            (len(bigg_met1) == 1) & (len(bigg_met2) == 0)
            | (len(bigg_met1) == 0) & (len(bigg_met2) == 1)
        ) & (id_r.startswith("EX_")):
            react_struct_checked.update(
                {id_r: StructuralR({id_r: "not_found_but_exchange_reaction"}, sel)}
            )
        else:
            react_struct_checked.update(
                {
                    id_r: StructuralR(
                        {"NOT_found": f"{id_r} not_pass_id_and_structural_checking"},
                        sel,
                    )
                }
            )
    return react_struct_checked


def runStructuralConversion(
    model_db: str,
    first_stage_selected_r: dict,
    first_stage_selected_m: dict,
    model: cobra.core.model.Model,
    bigg_network: dict,
    models_periplasmic: bool,
):
    """ Running structural conversion for all reactions. Selection reactions that have only 1 id as result """
    if model_db == "bigg":
        structural_conversion_r = runStructuralCheck(
            first_stage_selected_r, first_stage_selected_m, model, bigg_network
        )
    else:
        structural_conversion_r = {}
        for orig_id, selected in first_stage_selected_r.items():
            orig_met1 = [
                react.id for react in model.reactions.get_by_id(orig_id).reactants
            ]
            orig_met2 = [pro.id for pro in model.reactions.get_by_id(orig_id).products]
            if (len(orig_met1) > 20) | (len(orig_met2) > 20):
                structural_bigg_id = {
                    "Biomass": "No structural conversion since growth_reaction"
                }
            else:
                structural_bigg_id = convertReactionViaNetworkStructure(
                    orig_met1,
                    orig_met2,
                    first_stage_selected_m,
                    bigg_network,
                    models_periplasmic,
                )
            structural_conversion_r.update(
                {orig_id: StructuralR(structural_bigg_id, selected)}
            )
    return structural_conversion_r


def runSuggestionsMet(
    model_db: str, structural_rs: dict, struct_rs_sel: dict, met_selected: dict,
):
    # model uses bigg originally so there was no structural conversion, only checking
    if model_db == "bigg":
        structural_met = {
            orig_id: StructuralM(
                sel.highest_consistent,
                "No need for structural, because of bigg database",
                sel.compartments,
            )
            for orig_id, sel in met_selected.items()
        }
        sug_from_many = {}
    # actually getting suggestions based on reactions that were converted 1-1 with reaction equation
    else:
        structural_met = {}
        sug_to_many = defaultdict(list)
        sug_from_many = defaultdict(list)
        for r_old_id, r_sel in struct_rs_sel.items():
            if (
                r_sel.to_one_id == True
                and r_sel.from_one_id == True
                and structural_rs[r_old_id].comment.startswith(
                    "Found_via_one_to_many_metabolites"
                )
            ):
                for i in range(len(structural_rs[r_old_id].suggestions["orig_m"])):
                    sug_to_many[
                        structural_rs[r_old_id].suggestions["orig_m"][i]
                    ].append(structural_rs[r_old_id].suggestions["b_m"][i])
            elif structural_rs[r_old_id].comment.startswith(
                (
                    "Potentially_found_via_many_to_one_metabolites",
                    "Not_found_via_many_to_one_metabolites",
                    "Potentially_found_but_no_confidence",
                    "Not_all_metabolites_are_converted_but_important_for_many_original",
                )
            ):
                for ii in range(len(structural_rs[r_old_id].suggestions["orig_m"])):
                    sug_from_many[
                        structural_rs[r_old_id].suggestions["orig_m"][ii]
                    ].append(structural_rs[r_old_id].suggestions["b_m"][ii])
        sug_to_many = {k: v[0] for k, v in sug_to_many.items() if len(set(v)) == 1}
        for m_orig_id, met_sel in met_selected.items():
            if met_sel.to_one_id == True and met_sel.from_one_id == True:
                structural_met.update(
                    {
                        m_orig_id: StructuralM(
                            met_sel.highest_consistent,
                            "Already 1-1 conversion",
                            met_sel.compartments,
                        )
                    }
                )
            elif met_sel.to_one_id == False and met_sel.from_one_id == True:
                if m_orig_id in sug_to_many.keys():
                    structural_met.update(
                        {
                            m_orig_id: StructuralM(
                                [sug_to_many[m_orig_id]],
                                "Suggestion from one to many options",
                                met_sel.compartments,
                            )
                        }
                    )
                else:
                    structural_met.update(
                        {
                            m_orig_id: StructuralM(
                                [],
                                "No suggestion from one to many options",
                                met_sel.compartments,
                            )
                        }
                    )
            elif met_sel.to_one_id == True and met_sel.from_one_id == False:
                # checking how the same original id was converted in other models with different db
                # no information about the id in models with other db
                if not met_sel.in_other_models:
                    in_others = False
                else:
                    # checking whether in models with other db the id is converted as 1-1 ([True, True])
                    in_others = True
                    for val in met_sel.in_other_models.values():
                        if val != [True, True]:
                            in_others = False
                    # if the id is 1-1 in other models, checking whether other ids from the model, group that lead to n-1, are not 1-1 in models with other db
                    if in_others:
                        for oth_id in met_sel.from_many_other_ids:
                            for othval in met_selected[oth_id].in_other_models.values():
                                if othval == [True, True]:
                                    in_others = False
                if in_others:
                    structural_met.update(
                        {
                            m_orig_id: StructuralM(
                                met_sel.highest_consistent,
                                "Suggestion from many original ids to one bigg based on unique conversion from other models",
                                met_sel.compartments,
                            )
                        }
                    )
                # checking whether usage of conversion for one original ids from the group leads to better reaction equation conversion
                else:
                    in_r_eq = False
                    if m_orig_id in sug_from_many.keys():
                        if ("not_fit" not in set(sug_from_many[m_orig_id])) and (
                            {"no_data"} != set(sug_from_many[m_orig_id])
                        ):
                            in_r_eq = True
                            for ot_id in met_sel.from_many_other_ids:
                                if not {"not_fit", "no_data"} >= set(
                                    sug_from_many[ot_id]
                                ):
                                    in_r_eq = False
                    if in_r_eq:
                        structural_met.update(
                            {
                                m_orig_id: StructuralM(
                                    met_sel.highest_consistent,
                                    "Suggestion from many original ids to one bigg based on better reaction equation conversions for that original id in a group",
                                    met_sel.compartments,
                                )
                            }
                        )
                    else:
                        structural_met.update(
                            {
                                m_orig_id: StructuralM(
                                    [],
                                    "No suggestion from many original ids to one bigg",
                                    met_sel.compartments,
                                )
                            }
                        )
            elif met_sel.to_one_id == False and met_sel.from_one_id == False:
                structural_met.update(
                    {
                        m_orig_id: StructuralM(
                            [],
                            "No suggestion from many original ids to many bigg ids",
                            met_sel.compartments,
                        )
                    }
                )
            else:
                structural_met.update(
                    {
                        m_orig_id: StructuralM(
                            [],
                            "No suggestion from not converted",
                            met_sel.compartments,
                        )
                    }
                )
    return structural_met, sug_from_many



from collections import defaultdict


class Selected(object):
    """
    Class for selected object that checks results of different stages. After
    conversion stage replace_with_consistent = True and results, going further
    (highest_consistent), have to be consistent. After structural stage
    replace_with_consistent = False and results, going further
    (highest_consistent), don't change (highest). How results of consistency
    check differ from original is written in consistency attr. Compartments are
    just passed through for later. To_one_id shows amount of result ids
    (highest_consistent): True if 1, False if more than 1 and None if there is
    highest_consistent is empty. From_one_id is set as None and will become
    True or False in checkFromOneFromMany function. From_many_other_ids is set
    as empty list and will be populated later in checkFromOneFromMany function.
    """

    def __init__(
        self,
        compartments: list,
        replace_with_consistent: bool,
        highest: list = None,
        consistent: list = None,
        other_present=None,
        other_highest=None,
    ):
        if highest is None:
            highest = []
        if consistent is None:
            consistent = highest

        self.to_one_id = None
        self.from_one_id = None
        self.from_many_other_ids = []
        self.compartments = compartments
        if other_present is None:
            self.in_other_models = {}
        else:
            self.in_other_models = {other: [] for other in other_present}
        if replace_with_consistent:
            self.highest_consistent = consistent
        else:
            self.highest_consistent = highest
        if not consistent:
            if not highest:
                self.consistent = f"Not converted"
            elif replace_with_consistent:
                self.consistent = (
                    f"Not consistent. Originally bigg ids were {' '.join(highest)}. But in other models "
                    f"bigg ids were {other_highest}"
                )
            else:
                self.consistent = (
                    f"Can be considered as not consistent. Originally bigg ids were {' '.join(highest)}. "
                    f"But in other models bigg ids were {other_highest}"
                )
        elif set(consistent) == set(highest):
            self.consistent = "Yes"
        elif replace_with_consistent:
            self.consistent = (
                f"Changed: from {', '.join(highest)} to {', '.join(consistent)}. In other models bigg "
                f"ids were {other_highest}"
            )
        else:
            self.consistent = (
                f"Could be Changed: from {', '.join(highest)} to {', '.join(consistent)}. In other "
                f"models bigg ids were {other_highest}"
            )
        if len(self.highest_consistent) == 1:
            self.to_one_id = True
        if len(self.highest_consistent) > 1:
            self.to_one_id = False

    # adding results of mapping for selected objects with the same original id,
    # but from other models
    def check_others(self, original_id, selected: dict):
        for other_model in self.in_other_models.keys():
            self.in_other_models[other_model] = [
                selected[other_model][original_id].to_one_id,
                selected[other_model][original_id].from_one_id,
            ]


def checkDBConsistency(
    models_same_db: dict,
    converted_models: dict,
    attr_to_check: str,
    replace_with_consistent: bool,
):
    """
    Checking ID for different models with IDs from the same database. If one
    original ID from one database is converted separately for different models,
    looking for intersection of those lists. If no intersection found,
    conversion is inconsistent. If intersection is less that whole converted
    list, remove ids outside intersection.
    """
    consistent = defaultdict(dict)
    for db_type, models in models_same_db.items():
        # models is a dictionary of model_id - model_type
        num_model_types = len(set(models.values()))

        # no consistency check
        if num_model_types == 1 or db_type == "bigg":
            for model_id in models.keys():
                consistent[model_id] = {
                    idd: Selected(
                        species.compartments,
                        replace_with_consistent,
                        getattr(species, attr_to_check),
                    )
                    for idd, species in converted_models.get(model_id).items()
                }
        # consistency check
        else:
            # Get a unique set of ids for metabolites or reactions from all
            # models associated with the same database
            all_species_ids = set.union(
                *[
                    set(species.keys())  # mb or rcts ids
                    for model_id, species in converted_models.items()
                    if model_id in models.keys()
                ]
            )

            # looping through all id through all models with the same db type
            for iid in all_species_ids:
                # list for intersecting present ids from models with different
                # db #[[b1, b2], [b1]]
                bigg_ids = []
                # dict connecting ids and models for making .in_other_models attr
                bigg_ids_dict = {}
                models_present = []
                for model_id in models.keys():
                    species = converted_models.get(model_id).get(iid)
                    if species is not None:
                        species_attr = getattr(species, attr_to_check)

                        # collecting in models ids is present
                        models_present.append(model_id)

                        # adding only not empty lists because intersecting
                        # #[[b1, b2], []] isn't reasonable
                        if species_attr:
                            bigg_ids.append(species_attr)
                            bigg_ids_dict[model_id] = species_attr

                # if list for intersection is totally empty no intersection is needed
                if not bigg_ids:
                    for pres in models_present:
                        consistent[pres].update(
                            {
                                iid: Selected(
                                    converted_models.get(pres).get(iid).compartments,
                                    replace_with_consistent,
                                    other_present=set(models_present) - {pres},
                                )
                            }
                        )
                # intersecting result ids for the same original id from models with different db
                else:
                    common_ids = list(set.intersection(*map(set, bigg_ids)))
                    for present in models_present:
                        other_ids = {
                            k: v for k, v in bigg_ids_dict.items() if k != present
                        }
                        consistent[present].update(
                            {
                                iid: Selected(
                                    converted_models.get(present).get(iid).compartments,
                                    replace_with_consistent,
                                    getattr(
                                        converted_models.get(present).get(iid),
                                        attr_to_check,
                                    ),
                                    common_ids,
                                    set(models_present) - {present},
                                    other_ids,
                                )
                            }
                        )
    return consistent


def checkFromOneFromMany(selected: dict):
    """
    Checking in selected objects which original id where converted uniquely to
    one or several bigg ids (from_one_id = True) and which original ids have
    the same results (from_one_id = False).  Writing down original ids with the
    same results in from_many_other_ids.
    """
    # making dict with original ids and bigg results ids as str instead of list
    to_smth_str = {
        orig_id: " ".join(sorted(bigg_ids.highest_consistent))
        for orig_id, bigg_ids in selected.items()
    }
    # creating empty dict with bigg results as keys
    reversed_to_smth = {bigg_id: [] for bigg_id in to_smth_str.values()}
    # populating bigg keys in the empty dict with original ids, for which these bigg ids are results
    for k, v in to_smth_str.items():
        reversed_to_smth[v].append(k)
    # writing check results depending on amount of original ids for bigg results
    for value in reversed_to_smth.values():
        if len(value) == 1:
            selected[value[0]].from_one_id = True
        else:
            for val in value:
                selected[val].from_one_id = False
                selected[val].from_many_other_ids = list(set(value) - {val})


def run_selection(
    same_db_models: dict,
    previous_stage: dict,
    attr_to_check: str,
    replace_with_consistent=True,
):
    """
    Wrapping function to run checking consistency in resulting bigg ids for the
    same original id but different models with different DB. Then checking
    whether for particular model there is only one original id gives particular
    results or there are several original ids with the same results. And
    checking how the same original id performs in different models, can it give
    1 to 1 conversion somewhere else.If there is a group of original ids with
    the same results, connecting them together.
    """
    first_stage_selected = checkDBConsistency(
        same_db_models, previous_stage, attr_to_check, replace_with_consistent
    )
    for s in first_stage_selected.values():
        checkFromOneFromMany(s)
    for ss in first_stage_selected.values():
        for or_id, sel in ss.items():
            sel.check_others(or_id, first_stage_selected)
    return first_stage_selected



import gzip
import os
import re
import time
import warnings
from functools import partial
from mimetypes import guess_type
from pathlib import PosixPath

import ncbi_genome_download as ngd
import pandas as pd
from cobra import Model
from sympy import expand, symbols, sympify

from .general import is_float


def check_nt_or_aa(path_fasta: PosixPath):
    """Check whether fasta file is nt or aa.
    Codes are taken from  https://web.cas.org/help/BLAST/topics/codes.htm"""
    nt_letters = [
        "A",
        "C",
        "G",
        "T",
        "U",
        "R",
        "Y",
        "K",
        "M",
        "S",
        "W",
        "B",
        "D",
        "H",
        "V",
        "N",
        "-",
    ]
    aa_letters = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
        "K",
        "L",
        "M",
        "N",
        "O",
        "P",
        "Q",
        "R",
        "S",
        "T",
        "U",
        "V",
        "W",
        "X",
        "Y",
        "Z",
        "*",
        "-",
    ]
    aa_specific = list(set(aa_letters) - set(nt_letters))
    with open(path_fasta) as fasta_file:
        file_lines = fasta_file.readlines()
    sequence = ""
    for line in file_lines:
        if not line.startswith(">"):
            sequence = sequence + line.strip()
    aa_status = False
    for aa in aa_specific:
        if aa in sequence:
            aa_status = True
    return aa_status


def get_genome(ncbi_genome_name: str):
    encoding = guess_type(ncbi_genome_name)[1]  # uses file extension
    _open = partial(gzip.open, mode="rt") if encoding == "gzip" else open

    with _open(ncbi_genome_name) as input_fasta:
        genome_lines = input_fasta.readlines()
    genomes = {}
    genomeid = ""
    for line in genome_lines:
        if line.startswith(">"):
            if genomeid != "":
                genomes[genomeid] = genome
            genomeid = line.split()[0][1:]
            genome = ""
        else:
            genome = genome + line.strip()
    genomes[genomeid] = genome
    return genomes


def get_genes_gapseq(
    output_gene_folder: PosixPath,
    input_gapseq_genome_name: str,
    gapseq_model: Model,
    model_type: str,
    model_id: str,
):
    genomes = get_genome(input_gapseq_genome_name)
    head, tail = os.path.split(input_gapseq_genome_name)
    output_genes_name_gapseq = (
        output_gene_folder
        / f"{os.path.splitext(tail)[0]}_{model_type}_{model_id}_genes.faa"
    )
    gene_gapseq_fasta = open(output_genes_name_gapseq, "w")
    for gene in gapseq_model.genes:
        gene_gapseq_fasta.write(">" + gene.id + "\n")
        start = gene.id.split("_")[-2]
        end = gene.id.split("_")[-1]
        genomeid = gene.id.removeprefix("gp_").removesuffix("_" + start + "_" + end)
        if genomeid not in genomes.keys():
            genomeid = ".".join(genomeid.rsplit("_", 1))
        if genomeid not in genomes.keys():
            return False, False
        start = int(start)
        end = int(end)
        if start < end:
            gene_gapseq_fasta.write(genomes.get(genomeid)[start:end] + "\n")
        else:
            gene_gapseq_fasta.write(genomes.get(genomeid)[end:start] + "\n")
    gene_gapseq_fasta.close()
    aa_status = check_nt_or_aa(output_genes_name_gapseq)
    return output_genes_name_gapseq, aa_status


def get_genes_not_gapseq(
    output_gene_folder: PosixPath,
    input_genes_name: PosixPath,
    model: Model,
    model_type: str,
    model_id: str,
):
    encoding = guess_type(input_genes_name)[1]  # uses file extension
    _open = partial(gzip.open, mode="rt") if encoding == "gzip" else open

    with _open(input_genes_name) as input_fasta:
        lines = input_fasta.readlines()
    head, tail = os.path.split(input_genes_name)
    output_genes_name = (
        output_gene_folder
        / f"{os.path.splitext(tail)[0]}_{model_type}_{model_id}_genes.faa"
    )

    genes_fasta = open(output_genes_name, "w")
    new_id = ""
    for line in lines:
        if line.startswith(">"):
            if new_id != "":
                if new_id in model.genes:
                    genes_fasta.write(">" + new_id + "\n")
                    genes_fasta.write(gene)
            old_id = line.strip().split(" ")[0][1:]
            if model_type == "carveme":
                new_id = "_".join(old_id.rsplit(".", 1))
                new_id = new_id.replace(":", "_")
            elif model_type == "agora":
                new_id = old_id.removeprefix("fig|")
            elif model_type == "modelseed":
                new_id = old_id
            else:
                new_id = old_id
            gene = ""
        else:
            gene = gene + line
    if new_id in model.genes:
        genes_fasta.write(">" + new_id + "\n")
        genes_fasta.write(gene)
    genes_fasta.close()
    aa_status = check_nt_or_aa(output_genes_name)
    return output_genes_name, aa_status


def get_locus_tag_genes(
    ncbi_cds_name: str,
    ncbi_protein_name: str,
    feature_table_name: str,
    out_nt_fasta: str,
    out_aa_fasta: str,
    do_old_locus_tag: bool,
):
    feature_table = pd.read_csv(feature_table_name, sep="\t", compression="gzip")
    out_nt = open(out_nt_fasta, "w")
    with gzip.open(ncbi_cds_name, "rt") as cds:
        locus_tags = []
        for cdsl in cds:
            if cdsl.startswith(">"):
                if locus_tags != []:
                    for lt in locus_tags:
                        out_nt.write(">" + lt + "\n")
                        out_nt.write(one_seq)
                new_locus_tag = cdsl.split("[locus_tag=")[1].split("]")[0]
                if do_old_locus_tag:
                    attr = feature_table[
                        (feature_table["locus_tag"] == new_locus_tag)
                        & (feature_table["# feature"] == "gene")
                    ]["attributes"]
                    if attr.empty:
                        old_locus_tag = new_locus_tag
                    elif type(attr.values[0]) != str:
                        old_locus_tag = new_locus_tag
                    elif "old_locus_tag" not in attr.values[0]:
                        old_locus_tag = new_locus_tag
                    else:
                        old_locus_tag = attr.values[0].split("old_locus_tag=")[1]
                    locus_tags = old_locus_tag.split(",")
                else:
                    locus_tags = [new_locus_tag]
                one_seq = ""
            else:
                one_seq = one_seq + cdsl
    for lt in locus_tags:
        out_nt.write(">" + lt + "\n")
        out_nt.write(one_seq)
    out_nt.close()
    out_aa = open(out_aa_fasta, "w")
    with gzip.open(ncbi_protein_name, "rt") as proteins:
        locus_tags = []
        for protl in proteins:
            if protl.startswith(">"):
                if locus_tags != []:
                    for lt in locus_tags:
                        out_aa.write(">" + lt + "\n")
                        out_aa.write(one_seq)
                pp_id = protl.split(" ")[0][1:]
                pnew_locus_tag = feature_table[
                    (feature_table["product_accession"] == pp_id)
                    & (feature_table["# feature"] == "CDS")
                ]["locus_tag"].values[0]
                if do_old_locus_tag:
                    attr = feature_table[
                        (feature_table["locus_tag"] == pnew_locus_tag)
                        & (feature_table["# feature"] == "gene")
                    ]["attributes"]
                    if attr.empty:
                        pold_locus_tag = pnew_locus_tag
                    elif type(attr.values[0]) != str:
                        pold_locus_tag = pnew_locus_tag
                    elif "old_locus_tag" not in attr.values[0]:
                        pold_locus_tag = pnew_locus_tag
                    else:
                        pold_locus_tag = attr.values[0].split("old_locus_tag=")[1]
                    locus_tags = pold_locus_tag.split(",")
                else:
                    locus_tags = pnew_locus_tag.split(",")
                one_seq = ""
            else:
                one_seq = one_seq + protl
        for lt in locus_tags:
            out_aa.write(">" + lt + "\n")
            out_aa.write(one_seq)
    out_aa.close()


def get_final_fasta_with_ncbi_assemble(
    output_folder: PosixPath, assembly_id: str, do_old_locus_tag=True
):
    gene_path = output_folder / "tmp_gene_conversion"
    path = gene_path / "ncbi_assembly"
    path.mkdir(parents=True, exist_ok=True)
    ngd.download(
        assembly_accessions=assembly_id,
        output=path,
        file_formats="cds-fasta,protein-fasta,features",
        flat_output=True,
    )
    if not os.listdir(path):
        warnings.warn("\nWarning! NCBI download failed. Trying second time")
        time.sleep(15)
        ngd.download(
            assembly_accessions=assembly_id,
            output=path,
            file_formats="cds-fasta,protein-fasta,features",
            flat_output=True,
        )
    if not os.listdir(path):
        warnings.warn("\nWarning! NCBI download failed second time. Trying third time")
        time.sleep(15)
        ngd.download(
            assembly_accessions=assembly_id,
            output=path,
            file_formats="cds-fasta,protein-fasta,features",
            flat_output=True,
        )
    cds_faa = ""
    prot_faa = ""
    feat = ""
    for file in os.listdir(path):
        if file.endswith("_cds_from_genomic.fna.gz"):
            cds_faa = os.path.join(path, file)
        if file.endswith("_protein.faa.gz"):
            prot_faa = os.path.join(path, file)
        if file.endswith("_feature_table.txt.gz"):
            feat = os.path.join(path, file)
    if not cds_faa:
        raise ValueError("CDS fasta file is not found")
    if not prot_faa:
        raise ValueError("Protein fasta file is not found")
    if not feat:
        raise ValueError("Feature table is not found")
    final_nt_faa = os.path.join(output_folder, f"final_nt_{assembly_id}.faa")
    final_aa_faa = os.path.join(output_folder, f"final_aa_{assembly_id}.faa")
    get_locus_tag_genes(
        cds_faa,
        prot_faa,
        feat,
        final_nt_faa,
        final_aa_faa,
        do_old_locus_tag=do_old_locus_tag,
    )
    return final_nt_faa, final_aa_faa


def makeNewGPR(gpr: str, g_id_convert: dict):
    new_gpr = gpr
    mix_gpr = gpr
    for old_id, new_id in g_id_convert.items():
        new_id = new_id.replace(".", "_").replace('"', "").replace(":", "_")
        old_id = old_id.replace(".", "_").replace('"', "").replace(":", "_")
        new_gpr = new_gpr.replace(".", "_").replace('"', "").replace(":", "_")
        mix_gpr = mix_gpr.replace(".", "_").replace('"', "").replace(":", "_")
        if new_id[0].isdigit():
            new_id = "g_" + new_id

        new_gpr = re.sub(rf"\b{old_id}\b", new_id, new_gpr)

        if new_id != "not_found":
            mix_gpr = re.sub(rf"\b{old_id}\b", new_id, mix_gpr)
            globals()[new_id] = symbols(new_id)
        else:
            if old_id[0].isdigit():
                old_if_formated = "g_" + old_id
                mix_gpr = re.sub(rf"\b{old_id}\b", old_if_formated, mix_gpr)
                globals()[old_if_formated] = symbols(old_if_formated)
            else:
                globals()[old_id] = symbols(old_id)
    new_gpr = re.sub(r"\bnot_found\b", "1", new_gpr)
    new_gpr = re.sub(r"\bor\b", "+", new_gpr)
    new_gpr = re.sub(r"\band\b", "*", new_gpr)
    mix_gpr = re.sub(r"\bor\b", "+", mix_gpr)
    mix_gpr = re.sub(r"\band\b", "*", mix_gpr)
    equation = sympify(new_gpr)
    equation_expanded = expand(equation)
    equation_mix = sympify(mix_gpr)
    equation_expanded_mix = expand(equation_mix)
    if is_float(str(equation_expanded)):
        new_gpr_expanded = ""
    else:
        new_gpr_expanded = str(equation_expanded).split(" + ")
        for i in range(len(new_gpr_expanded)):
            genes_and = []
            for g in new_gpr_expanded[i].split("*"):
                if g and (not is_float(g)):
                    genes_and.append(g)
            if not genes_and:
                new_gpr_expanded.pop(i)
            elif (len(genes_and) > 1) and (len(new_gpr_expanded) > 1):
                new_gpr_expanded[i] = "(" + " and ".join(sorted(genes_and)) + ")"
            else:
                new_gpr_expanded[i] = " and ".join(sorted(genes_and))
        new_gpr_expanded = " or ".join(sorted(new_gpr_expanded))
    new_gpr_expanded_mix = str(equation_expanded_mix).split(" + ")
    for j in range(len(new_gpr_expanded_mix)):
        genes_and_mix = []
        for gm in new_gpr_expanded_mix[j].split("*"):
            if gm and (not is_float(gm)):
                genes_and_mix.append(gm)
        if not genes_and_mix:
            new_gpr_expanded_mix.pop(j)
        elif (len(genes_and_mix) > 1) and (len(new_gpr_expanded_mix) > 1):
            new_gpr_expanded_mix[j] = "(" + " and ".join(sorted(genes_and_mix)) + ")"
        else:
            new_gpr_expanded_mix[j] = " and ".join(sorted(genes_and_mix))
    new_gpr_expanded_mix = " or ".join(sorted(new_gpr_expanded_mix))
    return new_gpr_expanded, new_gpr_expanded_mix


def uniteGPR(list_gpr: [str]):
    united_and = []
    for gpr in list_gpr:
        for genes_ands in gpr.split(" or "):
            united_and.append(genes_ands)
    unique_and = list(set(united_and))
    if len(unique_and) > 1:
        for i in range(len(unique_and)):
            if ("and" in unique_and[i]) and ("(" not in unique_and[i]):
                unique_and[i] = "(" + unique_and[i] + ")"
    united_gpr = " or ".join(sorted(unique_and))
    return united_gpr



import hashlib
import logging
import os
import pickle
import subprocess
import sys
import warnings
from collections import defaultdict
from copy import deepcopy
from pathlib import Path

from cobra.io import read_sbml_model
from platformdirs import user_data_path

from .conversion import (
    ConvAgora,
    ConvBase,
    ConvBigg,
    ConvCarveme,
    ConvGapseq,
    ConvMetanetx,
    ConvModelseed,
    no_changes_for_notconv,
    remove_zero_for_notconv,
    replace_square_brackets,
)
from .creation import SuperModel
from .curation import get_duplicated_reactions, remove_b_type_exchange
from .dbs import download_db, get_bigg_network
from .genes import (
    get_final_fasta_with_ncbi_assemble,
    get_genes_gapseq,
    get_genes_not_gapseq,
)
from .periplasmic import getSuggestionPeriplasmic
from .selection import run_selection
from .structural import runStructuralConversion, runSuggestionsMet


def get_env():
    """
    Function to appen additional path to system PATH env var and return dict
    with new env. In case the package is used in conda env.
    """
    add_bin = str(Path(sys.executable).parent)
    tmp_env = os.environ.copy()
    tmp_env["PATH"] = f"{tmp_env['PATH']}:{add_bin}"
    return tmp_env


def sha256sum(filename):
    h = hashlib.sha256()
    b = bytearray(128 * 1024)
    mv = memoryview(b)
    with open(filename, "rb", buffering=0) as f:
        while n := f.readinto(mv):
            h.update(mv[:n])
    return h.hexdigest()


class LoggerContext:
    def __init__(self, logger_name, show_logs=False):
        self.__show_logs__ = show_logs
        self.__logger__ = logging.getLogger(logger_name)

    def __enter__(self):
        if not self.__show_logs__:
            self.__logger__.setLevel(logging.CRITICAL)

    def __exit__(self, exit_type, exit_value, exit_traceback):
        if not self.__show_logs__:
            self.__logger__.setLevel(logging.NOTSET)


def load_sbml_model(path_to_model, cache: bool = True, show_logs: bool = False):

    file_hash = sha256sum(path_to_model)
    path_to_model = Path(path_to_model)
    cache_path = path_to_model.parent / (path_to_model.stem + "_" + file_hash + ".pkl")

    if cache_path.exists():
        with open(cache_path, "rb") as fh:
            model = pickle.load(fh)
    else:
        # Read the cobra model
        with LoggerContext("cobra", show_logs):
            model = read_sbml_model(path_to_model)
        # Cache the model to a pickle if option is set
        if cache:
            with open(cache_path, "wb") as fh:
                pickle.dump(model, fh)
    return model


class GatheredModels:
    """
    Class that gathers information and necessary conversion results for all
    models. Input for the class and tool in general is dictionary with all
    models and related information.

    Parameters
    ----------
    dict_of_all_models_with_feature : dict
        Dictionary of the following format:
        {
            model_id:
            {
                'path_to_model': str,
                'model_type': str one of (agora, carveme, gapseq, modelseed) or custom,
                'path_to_genome': str (can be '' or None if convert_genes = False)
            }
        }
        Note: if custom `model_type` create type class in advance,
    assembly : None, optional
    path_final_genome_nt : None, optional
    path_final_genome_aa : None, optional
    custom_model_type : None, optional

    Notes
    -----
    If all 3 parameters bellow are None then gene conversion is not done and
    genomes for model_id are not need.
    """

    def __init__(
        self,
        custom_model_type=None,
        clear_db_cache=False,
    ):
        # If specified, clear the cached conversion tables and dictionaries
        if clear_db_cache:
            for p in user_data_path("gemsembler").iterdir():
                p.unlink()

        self.__conf = {
            "agora": {
                "remove_b": False,
                "db_name": "weird_bigg",
                "wo_periplasmic": True,
                "conv_strategy": ConvAgora(),
                "genome_model_strategy": get_genes_not_gapseq,
                "alter_notconv_m": replace_square_brackets,
                "alter_notconv_r": no_changes_for_notconv,
            },
            "carveme": {
                "remove_b": False,
                "db_name": "bigg",
                "wo_periplasmic": False,
                "conv_strategy": ConvCarveme(),
                "genome_model_strategy": get_genes_not_gapseq,
                "alter_notconv_m": no_changes_for_notconv,
                "alter_notconv_r": no_changes_for_notconv,
            },
            "bigg": {
                "remove_b": False,
                "db_name": "bigg",
                "wo_periplasmic": False,
                "conv_strategy": ConvBigg(),
                "genome_model_strategy": get_genes_not_gapseq,
                "alter_notconv_m": no_changes_for_notconv,
                "alter_notconv_r": no_changes_for_notconv,
            },
            "metanetx": {
                "remove_b": False,
                "db_name": "bigg",
                "wo_periplasmic": ConvMetanetx(),
                "genome_model_strategy": get_genes_not_gapseq,
                "alter_notconv_m": no_changes_for_notconv,
                "alter_notconv_r": no_changes_for_notconv,
            },
            "gapseq": {
                "remove_b": False,
                "db_name": "modelseed",
                "wo_periplasmic": False,
                "conv_strategy": ConvGapseq(),
                "genome_model_strategy": get_genes_gapseq,
                "alter_notconv_m": remove_zero_for_notconv,
                "alter_notconv_r": remove_zero_for_notconv,
            },
            "modelseed": {
                "remove_b": True,
                "db_name": "modelseed",
                "wo_periplasmic": True,
                "conv_strategy": ConvModelseed(),
                "genome_model_strategy": get_genes_not_gapseq,
                "alter_notconv_m": remove_zero_for_notconv,
                "alter_notconv_r": remove_zero_for_notconv,
            },
        }

        self.__models = {}
        self.first_stage_selected_metabolites = None
        self.first_stage_selected_reactions = None
        self.structural_first_run_reactions = defaultdict(dict)
        self.second_stage_selected_reactions = None
        self.structural_first_run_metabolites = defaultdict(dict)
        self.second_stage_selected_metabolites = None
        self.structural_second_run_reactions = defaultdict(dict)
        self.third_stage_selected_reactions = None
        self.many_to_one_sug = defaultdict(dict)
        self.periplasmic_metabolites = defaultdict(dict)
        self.periplasmic_reactions = defaultdict(dict)

    def __contains__(self, item):
        return item in self.__models

    def __len__(self):
        return len(self.__models)

    def get_conf(self, model_type=None):
        if model_type is None:
            return deepcopy(self.__conf)
        else:
            return deepcopy(self.__conf.get(model_type))

    def get_model_attrs(self, model_id=None, attr=None):
        if model_id is not None and attr is not None:
            return deepcopy(self.__models.get(model_id).get(attr))
        elif model_id is not None and attr is None:
            return deepcopy(self.__models.get(model_id))
        else:
            return deepcopy(self.__models)

    @property
    def same_db_models(self):
        same_db_models = defaultdict(dict)
        for model_id, model_attrs in self.__models.items():
            model_type = model_attrs["model_type"]
            db_name = self.__conf.get(model_type).get("db_name")
            same_db_models[db_name][model_id] = model_type
        return same_db_models

    @property
    def converted_metabolites(self):
        conv_mbs = defaultdict(dict)
        for model_id, model_attrs in self.__models.items():
            converter = self.__conf.get(model_attrs["model_type"]).get("conv_strategy")
            conv_mbs[model_id] = converter.convert_all_metabolites(
                model_attrs["preprocess_model"]
            )
        return conv_mbs

    @property
    def converted_reactions(self):
        conv_rcts = defaultdict(dict)
        for model_id, model_attrs in self.__models.items():
            converter = self.__conf.get(model_attrs["model_type"]).get("conv_strategy")
            conv_rcts[model_id] = converter.convert_all_reactions(
                model_attrs["preprocess_model"]
            )
        return conv_rcts

    def run(self):
        # run first convertion
        print("Running initial convertion")
        self.first_stage_selected_metabolites = run_selection(
            self.same_db_models, self.converted_metabolites, "highest"
        )
        self.first_stage_selected_reactions = run_selection(
            self.same_db_models, self.converted_reactions, "highest"
        )

        # run first structural conversion
        print("Running 1st structural convertion")
        bigg_network = get_bigg_network()
        for model_id, first_sel in self.first_stage_selected_reactions.items():
            model_type = self.__models[model_id]["model_type"]
            db_name = self.__conf.get(model_type).get("db_name")
            self.structural_first_run_reactions[model_id] = runStructuralConversion(
                db_name,
                first_sel,
                self.first_stage_selected_metabolites[model_id],
                self.__models[model_id]["preprocess_model"],
                bigg_network,
                False,
            )
        # run second stage selection for first structural reactions
        self.second_stage_selected_reactions = run_selection(
            self.same_db_models,
            self.structural_first_run_reactions,
            "structural",
            replace_with_consistent=False,
        )

        # get suggestions from structural reactions for metabolites
        print("Running structural suggestions for metabolites")
        for model_id, rs_struct_sel in self.second_stage_selected_reactions.items():
            model_type = self.__models[model_id]["model_type"]
            db_mod = self.__conf.get(model_type).get("db_name")
            (
                self.structural_first_run_metabolites[model_id],
                self.many_to_one_sug[model_id],
            ) = runSuggestionsMet(
                db_mod,
                self.structural_first_run_reactions[model_id],
                rs_struct_sel,
                self.first_stage_selected_metabolites[model_id],
            )
        # run second stage selection for suggestions for metabolites from structural
        self.second_stage_selected_metabolites = run_selection(
            self.same_db_models, self.structural_first_run_metabolites, "structural"
        )

        # run second structural conversion with suggestions for metabolites
        print("Running 2d structural convertion")
        for model_id, sec_sel in self.second_stage_selected_reactions.items():
            model_type = self.__models[model_id]["model_type"]
            db_name = self.__conf.get(model_type).get("db_name")
            self.structural_second_run_reactions[model_id] = runStructuralConversion(
                db_name,
                sec_sel,
                self.second_stage_selected_metabolites[model_id],
                self.__models[model_id]["preprocess_model"],
                bigg_network,
                self.__conf.get(model_type).get("wo_periplasmic"),
            )
        # run third stage selection for first structural reactions
        self.third_stage_selected_reactions = run_selection(
            self.same_db_models,
            self.structural_second_run_reactions,
            "structural",
            replace_with_consistent=False,
        )

        print("Introducing periplasmic compartment")
        # introducing periplasmic compartment for models, that don't have it originally
        for model_id, th_sel in self.third_stage_selected_reactions.items():
            if self.__conf.get(self.__models[model_id]["model_type"]).get(
                "wo_periplasmic"
            ):
                (
                    self.periplasmic_metabolites[model_id],
                    self.periplasmic_reactions[model_id],
                ) = getSuggestionPeriplasmic(
                    th_sel,
                    self.structural_second_run_reactions[model_id],
                    self.second_stage_selected_metabolites[model_id],
                    self.__models[model_id]["preprocess_model"],
                    bigg_network,
                )
            else:
                (
                    self.periplasmic_metabolites[model_id],
                    self.periplasmic_reactions[model_id],
                ) = ({}, {})

    def get_input_dictionaries(self):
        final_r_sel = defaultdict(dict)
        final_r_not_sel = defaultdict(dict)
        final_m_sel = defaultdict(dict)
        final_m_not_sel = defaultdict(dict)
        periplasmic_m = defaultdict(dict)
        periplasmic_r = defaultdict(dict)
        for model_id in self.__models.keys():
            final_r_sel[model_id] = {}
            final_r_not_sel[model_id] = {}
            final_m_sel[model_id] = {}
            final_m_not_sel[model_id] = {}
            for orig_r_id, sel_r in self.third_stage_selected_reactions[
                model_id
            ].items():
                if (sel_r.to_one_id == True and sel_r.from_one_id == True) or (
                    sel_r.to_one_id == True
                    and sel_r.from_one_id == False
                    and (
                        (
                            orig_r_id
                            in self.__models[model_id]["duplicated_reactions"][
                                "ID"
                            ].tolist()
                        )
                        or (sel_r.highest_consistent == ["Biomass"])
                    )
                ):
                    final_r_sel[model_id].update(
                        {orig_r_id: [sel_r.compartments, sel_r.highest_consistent]}
                    )
                else:
                    orig_r_id_alt = self.__conf[self.__models[model_id]["model_type"]][
                        "alter_notconv_r"
                    ](orig_r_id)
                    final_r_not_sel[model_id].update(
                        {orig_r_id: [sel_r.compartments, [orig_r_id_alt]]}
                    )
                if (
                    len(
                        self.__models[model_id]["preprocess_model"]
                        .reactions.get_by_id(orig_r_id)
                        .reactants
                    )
                    > 24
                ):
                    final_r_not_sel[model_id].update(
                        {orig_r_id: [sel_r.compartments, sel_r.highest_consistent]}
                    )
            for orig_m_id, sel_m in self.second_stage_selected_metabolites[
                model_id
            ].items():
                if orig_m_id in self.periplasmic_metabolites[model_id]:
                    if self.periplasmic_metabolites[model_id][orig_m_id].replace:
                        final_m_sel[model_id].update(
                            {
                                orig_m_id: [
                                    ["p"],
                                    [
                                        self.periplasmic_metabolites[model_id][
                                            orig_m_id
                                        ].bigg_p
                                    ],
                                ]
                            }
                        )
                    else:
                        final_m_sel[model_id].update(
                            {orig_m_id: [sel_m.compartments, sel_m.highest_consistent]}
                        )
                        periplasmic_m[model_id].update(
                            {
                                orig_m_id: [
                                    ["p"],
                                    [
                                        self.periplasmic_metabolites[model_id][
                                            orig_m_id
                                        ].bigg_p
                                    ],
                                ]
                            }
                        )
                else:
                    if sel_m.to_one_id == True and sel_m.from_one_id == True:
                        final_m_sel[model_id].update(
                            {orig_m_id: [sel_m.compartments, sel_m.highest_consistent]}
                        )
                    else:
                        orig_m_id_alt = self.__conf[
                            self.__models[model_id]["model_type"]
                        ]["alter_notconv_m"](orig_m_id)
                        final_m_not_sel[model_id].update(
                            {orig_m_id: [sel_m.compartments, [orig_m_id_alt]]}
                        )
            if self.periplasmic_reactions[model_id]:
                for p_r_id, p_r in self.periplasmic_reactions[model_id].items():
                    periplasmic_r[model_id].update({p_r_id: p_r.metabolites_changed_p})

        return (
            final_m_sel,
            final_m_not_sel,
            final_r_sel,
            final_r_not_sel,
            periplasmic_m,
            periplasmic_r,
        )

    def assemble_supermodel(
        self,
        output_folder: str,
        assembly_id=None,
        path_final_genome_nt=None,
        path_final_genome_aa=None,
        evalue_threshold=0.001,
        do_mix_conv_notconv=False,
        and_as_solid=False,
        do_old_locus_tag=True,
    ):
        # Check if assembly and final genome are present.
        # If not, throw a warning.
        if not assembly_id and not path_final_genome_nt and not path_final_genome_aa:
            warnings.warn(
                "\nWarning! No final genome for gene conversion is provided. "
                "Gene conversion will not be performed.\nIf you want to "
                "convert genes, please provide either assembly id or custom "
                "fasta files (nt/aa/both), \nto which genes must be converted."
            )
            gene_path = None
        elif assembly_id and (path_final_genome_nt or path_final_genome_aa):
            warnings.warn(
                "\nWarning! Both assembly and user final genome for gene conversion are provided. "
                "Gene conversion will not be performed.\nIf you want to "
                "convert genes, please provide one of both either assembly id or custom "
                "fasta files (nt/aa/both), to which genes must be converted."
            )
            gene_path = None
        else:
            # GGE: Check that BLAST is installed
            proc = subprocess.run(
                f"makeblastdb -h",
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            if proc.returncode != 0:
                raise OSError(
                    "Check that makeblastdb (and BLAST in general) is properly installed!"
                    "\n(a remainder, that you NEED to install blast to use this package)"
                )
            output_folder = Path(output_folder)
            gene_path = output_folder / "tmp_gene_conversion"
            gene_path.mkdir(exist_ok=True, parents=True)
            db_path = gene_path / "blast_db"
            db_path.mkdir(exist_ok=True, parents=True)
            print("Downloading assembly from NCBI")
            if assembly_id:
                (
                    path_final_genome_nt,
                    path_final_genome_aa,
                ) = get_final_fasta_with_ncbi_assemble(
                    output_folder, assembly_id, do_old_locus_tag=do_old_locus_tag
                )
            if path_final_genome_nt is not None:
                print("Building BLAST database")
                subprocess.run(
                    f"makeblastdb -in {path_final_genome_nt} -out "
                    f"{Path(db_path, 'nt_db')} -dbtype nucl"
                    f" -title nt_db -parse_seqids",  # WindowsFix ('' removed)  [seems to work on Linux as well]
                    shell=True,
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    env=get_env(),
                )
            if path_final_genome_aa is not None:
                print("Building BLAST database")
                subprocess.run(
                    f"makeblastdb -in {path_final_genome_aa} -out "
                    f"{Path(db_path, 'aa_db')} -dbtype"
                    f" prot -title aa_db -parse_seqids",  # WindowsFix ('' removed) [seems to work on Linux as well]
                    shell=True,
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    env=get_env(),
                )
            for model_id, model_data in self.__models.items():
                print(f"Running gene conversion with BLAST for {model_id}")
                if model_data["path_to_genome"] == "":
                    continue
                out_blast_file = gene_path / (model_id + "_blast.tsv")
                model_gene_file, aa_status = self.__conf[model_data["model_type"]][
                    "genome_model_strategy"
                ](
                    gene_path,
                    model_data["path_to_genome"],
                    model_data["preprocess_model"],
                    model_data["model_type"],
                    model_id,
                )
                blast_command = ""
                db_name = ""
                if aa_status and path_final_genome_aa is not None:
                    blast_command = "blastp"
                    db_name = "aa_db"  # WindowsFix ('' removed) [seems to work on Linux as well]
                elif aa_status and path_final_genome_nt is not None:
                    blast_command = "tblastn"
                    db_name = "nt_db"  # WindowsFix ('' removed) [seems to work on Linux as well]
                elif not aa_status and path_final_genome_nt is not None:
                    blast_command = "blastn"
                    db_name = "nt_db"  # WindowsFix ('' removed) [seems to work on Linux as well]
                elif not aa_status and path_final_genome_aa is not None:
                    blast_command = "blastx"
                    db_name = "aa_db"  # WindowsFix ('' removed) [seems to work on Linux as well]
                if blast_command == "" or db_name == "":
                    warnings.warn("\nWarning! Something wrong with aa/nt in files/DB")
                elif not model_gene_file:
                    warnings.warn("\nWarning! Something wrong with gene file")
                else:
                    subprocess.run(
                        f"{blast_command} -query {model_gene_file} "
                        f"-db {Path(db_path, db_name)} "
                        f"-max_target_seqs 1 -evalue {evalue_threshold} "
                        f"-outfmt 6 -out {out_blast_file}",  # WindowsFix ('' removed) [seems to work on Linux as well]
                        shell=True,
                        check=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        env=get_env(),
                    )
        print("Assembling Supermodel")
        # Get final tables to create new objects
        (
            final_m_sel,
            final_m_not_sel,
            final_r_sel,
            final_r_not_sel,
            periplasmic_m,
            periplasmic_r,
        ) = self.get_input_dictionaries()

        # Create supermodel
        bigg_data_m = download_db(
            "http://bigg.ucsd.edu/static/namespace/bigg_models_metabolites.txt",
            "bigg_models_metabolites.txt.gz",
        )
        bigg_data_r = download_db(
            "http://bigg.ucsd.edu/static/namespace/bigg_models_reactions.txt",
            "bigg_models_reactions.txt.gz",
        )
        supermodel = SuperModel(
            False,
            {
                "type": "SuperModel",
                "args": {
                    "final_m_sel": final_m_sel,
                    "final_m_not_sel": final_m_not_sel,
                    "final_r_sel": final_r_sel,
                    "final_r_not_sel": final_r_not_sel,
                    "all_models_data": self.__models,
                    "additional_periplasmic_m": periplasmic_m,
                    "periplasmic_r": periplasmic_r,
                    "m_db_info": bigg_data_m,
                    "r_db_info": bigg_data_r,
                    "gene_folder": gene_path,
                    "do_mix_conv_notconv": do_mix_conv_notconv,
                    "and_as_solid": and_as_solid,
                },
            },
        )
        return supermodel

    def set_configuration(
        self,
        model_type: str,
        remove_b: bool,
        db_name: str,
        wo_periplasmic: bool,
        conv_strategy,
        genome_model_strategy,
        **kwargs,
    ):
        assert isinstance(conv_strategy, ConvBase)

        # TODO: add checks on conf input arg
        self.__conf[model_type] = {
            "remove_b": remove_b,
            "db_name": db_name,
            "wo_periplasmic": wo_periplasmic,
            "conv_strategy": conv_strategy,
            "genome_model_strategy": genome_model_strategy,
            **kwargs,
        }

    def add_model(
        self,
        model_id: str,
        path_to_model: str,
        model_type: str,
        path_to_genome: str = None,
        cache: bool = True,
        show_logs: bool = False,
    ):
        # Run checks on model_id and model_type
        # TODO: check with conversion dictionaries
        assert model_id not in self.__models, f"model_id {model_id} already used"
        assert model_type in self.__conf, f"Missing configuration for {model_type}"

        model = load_sbml_model(path_to_model, cache, show_logs)

        # Populate the internal data
        self.__models[model_id] = {
            "original_model": deepcopy(model),
            "path_to_model": path_to_model,
            "model_type": model_type,
            "path_to_genome": path_to_genome,
        }

        # If model_type requires it, remove `_b` extensions
        if self.__conf.get(model_type).get("remove_b"):
            model = remove_b_type_exchange(model)
        self.__models[model_id]["preprocess_model"] = model

        dupl_r = get_duplicated_reactions(model)
        self.__models[model_id]["duplicated_reactions"] = dupl_r
